################################################################################
# Copyright (C) 2011-2013 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
General numerical functions and methods.

"""
from scipy.optimize import approx_fprime
import functools
import itertools
import operator

import sys
import getopt

import numpy as np
import scipy as sp
import scipy.linalg as linalg
import scipy.special as special
import scipy.optimize as optimize
import scipy.sparse as sparse

import tempfile as tmp

import unittest
from numpy import testing


def flatten_axes(X, *ndims):
    ndim = sum(ndims)
    if np.ndim(X) < ndim:
        raise ValueError("Not enough ndims in the array")
    if len(ndims) == 0:
        return X
    shape = np.shape(X)
    i = np.ndim(X) - ndim
    plates = shape[:i]
    nd_sums = i + np.cumsum((0,) + ndims)
    sizes = tuple(
        np.prod(shape[i:j])
        for (i, j) in zip(nd_sums[:-1], nd_sums[1:])
    )
    return np.reshape(X, plates + sizes)


def reshape_axes(X, *shapes):
    ndim = len(shapes)
    if np.ndim(X) < ndim:
        raise ValueError("Not enough ndims in the array")
    i = np.ndim(X) - ndim
    sizes = tuple(np.prod(sh) for sh in shapes)
    if np.shape(X)[i:] != sizes:
        raise ValueError("Shapes inconsistent with sizes")
    shape = tuple(i for sh in shapes for i in sh)
    return np.reshape(X, np.shape(X)[:i] + shape)


def find_set_index(index, set_lengths):
    """
    Given set sizes and an index, returns the index of the set

    The given index is for the concatenated list of the sets.
    """
    # Negative indices to positive
    if index < 0:
        index += np.sum(set_lengths)

    # Indices must be on range (0, N-1)
    if index >= np.sum(set_lengths) or index < 0:
        raise Exception("Index out bounds")

    return np.searchsorted(np.cumsum(set_lengths), index, side='right')


def parse_command_line_arguments(mandatory_args, *optional_args_list, argv=None):
    """
    Parse command line arguments of style "--parameter=value".

    Parameter specification is tuple: (name, converter, description).

    Some special handling:

    * If converter is None, the command line does not accept any value
      for it, but instead use either "--option" to enable or
      "--no-option" to disable.

    * If argument name contains hyphens, those are converted to
      underscores in the keys of the returned dictionaries.

    Parameters
    ----------

    mandatory_args : list of tuples
        Specs for mandatory arguments

    optional_args_list : list of lists of tuples
        Specs for each optional arguments set

    argv : list of strings (optional)
        The command line arguments. By default, read sys.argv.

    Returns
    -------

    args : dictionary
        The parsed mandatory arguments

    kwargs : dictionary
        The parsed optional arguments

    Examples
    --------

    >>> from pprint import pprint as print
    >>> from bayespy.utils import misc
    >>> (args, kwargs) = misc.parse_command_line_arguments(
    ...     # Mandatory arguments
    ...     [
    ...         ('name',     str,  "Full name"),
    ...         ('age',      int,  "Age (years)"),
    ...         ('employed', None, "Working"),
    ...     ],
    ...     # Optional arguments
    ...     [
    ...         ('phone',          str, "Phone number"),
    ...         ('favorite-color', str, "Favorite color")
    ...     ],
    ...     argv=['--name=John Doe',
    ...           '--age=42',
    ...           '--no-employed',
    ...           '--favorite-color=pink']
    ... )
    >>> print(args)
    {'age': 42, 'employed': False, 'name': 'John Doe'}
    >>> print(kwargs)
    {'favorite_color': 'pink'}

    It is possible to have several optional argument sets:

    >>> (args, kw_info, kw_fav) = misc.parse_command_line_arguments(
    ...     # Mandatory arguments
    ...     [
    ...         ('name',     str,  "Full name"),
    ...     ],
    ...     # Optional arguments (contact information)
    ...     [
    ...         ('phone', str, "Phone number"),
    ...         ('email', str, "E-mail address")
    ...     ],
    ...     # Optional arguments (preferences)
    ...     [
    ...         ('favorite-color', str, "Favorite color"),
    ...         ('favorite-food',  str, "Favorite food")
    ...     ],
    ...     argv=['--name=John Doe',
    ...           '--favorite-color=pink',
    ...           '--email=john.doe@email.com',
    ...           '--favorite-food=spaghetti']
    ... )
    >>> print(args)
    {'name': 'John Doe'}
    >>> print(kw_info)
    {'email': 'john.doe@email.com'}
    >>> print(kw_fav)
    {'favorite_color': 'pink', 'favorite_food': 'spaghetti'}

    """

    if argv is None:
        argv = sys.argv[1:]

    mandatory_arg_names = [arg[0] for arg in mandatory_args]

    # Sizes of each optional argument list
    optional_args_lengths = [len(opt_args) for opt_args in optional_args_list]

    all_args = mandatory_args + functools.reduce(operator.add, optional_args_list, [])

    # Create a list of arg names for the getopt parser
    arg_list = []
    for arg in all_args:
        arg_name = arg[0].lower()
        if arg[1] is None:
            arg_list.append(arg_name)
            arg_list.append("no-" + arg_name)
        else:
            arg_list.append(arg_name + "=")

    if len(set(arg_list)) < len(arg_list):
        raise Exception("Argument names are not unique")

    # Use getopt parser
    try:
        (cl_opts, cl_args) = getopt.getopt(argv, "", arg_list)
    except getopt.GetoptError as err:
        print(err)
        print("Usage:")
        for arg in all_args:
            if arg[1] is None:
                print("--{0}\t{1}".format(arg[0].lower(),
                                          arg[2]))
            else:
                print("--{0}=<{1}>\t{2}".format(arg[0].lower(),
                                               str(arg[1].__name__).upper(),
                                               arg[2]))
        sys.exit(2)

    # A list of all valid flag names: ["--first-argument", "--another-argument"]
    valid_flags = []
    valid_flag_arg_indices = []
    for (ind, arg) in enumerate(all_args):
        valid_flags.append("--" + arg[0].lower())
        valid_flag_arg_indices.append(ind)
        if arg[1] is None:
            valid_flags.append("--no-" + arg[0].lower())
            valid_flag_arg_indices.append(ind)

    # Go through all the given command line arguments and store them in the
    # correct dictionaries
    args = dict()
    kwargs_list = [dict() for i in range(len(optional_args_list))]
    handled_arg_names = []
    for (cl_opt, cl_arg) in cl_opts:

        # Get the index of the argument
        try:
            ind = valid_flag_arg_indices[valid_flags.index(cl_opt.lower())]
        except ValueError:
            print("Invalid command line argument: {0}".format(cl_opt))
            raise Exception("Invalid argument given")

        # Check that the argument wasn't already given and then mark the
        # argument as handled
        if all_args[ind][0] in handled_arg_names:
            raise Exception("Same argument given multiple times")
        else:
            handled_arg_names.append(all_args[ind][0])

        # Check whether to add the argument to the mandatory or optional
        # argument dictionary
        if ind < len(mandatory_args):
            dict_to = args
        else:
            dict_index = find_set_index(ind - len(mandatory_args),
                                        optional_args_lengths)
            dict_to = kwargs_list[dict_index]

        # Convert and store the argument
        convert_function = all_args[ind][1]
        arg_name = all_args[ind][0].replace('-', '_')
        if convert_function is None:
            if cl_opt[:5] == "--no-":
                dict_to[arg_name] = False
            else:
                dict_to[arg_name] = True
        else:
            dict_to[arg_name] = convert_function(cl_arg)

    # Check if some mandatory argument was not given
    for arg_name in mandatory_arg_names:
        if arg_name not in handled_arg_names:
            raise Exception("Mandatory argument --{0} not given".format(arg_name))

    return tuple([args] + kwargs_list)


def composite_function(function_list):
    """
    Construct a function composition from a list of functions.

    Given a list of functions [f,g,h], constructs a function :math:`h \circ g
    \circ f`.  That is, returns a function :math:`z`, for which :math:`z(x) =
    h(g(f(x)))`.
    """
    def composite(X):
        for function in function_list:
            X = function(X)
        return X
    return composite


def ceildiv(a, b):
    """
    Compute a divided by b and rounded up.
    """
    return -(-a // b)

def rmse(y1, y2, axis=None):
    return np.sqrt(np.mean((y1-y2)**2, axis=axis))

def is_callable(f):
    return hasattr(f, '__call__')

def atleast_nd(X, d):
    if np.ndim(X) < d:
        sh = (d-np.ndim(X))*(1,) + np.shape(X)
        X = np.reshape(X, sh)
    return X

def T(X):
    """
    Transpose the matrix.
    """
    return np.swapaxes(X, -1, -2)

class TestCase(unittest.TestCase):
    """
    Simple base class for unit testing.

    Adds NumPy's features to Python's unittest.
    """

    def assertAllClose(self, A, B,
                       msg="Arrays not almost equal",
                       rtol=1e-4,
                       atol=0):

        self.assertEqual(np.shape(A), np.shape(B), msg=msg)
        testing.assert_allclose(A, B, err_msg=msg, rtol=rtol, atol=atol)
        pass

    def assertArrayEqual(self, A, B, msg="Arrays not equal"):
        self.assertEqual(np.shape(A), np.shape(B), msg=msg)
        testing.assert_array_equal(A, B, err_msg=msg)
        pass

    def assertMessage(self, M1, M2):

        if len(M1) != len(M2):
            self.fail("Message lists have different lengths")

        for (m1, m2) in zip(M1, M2):
            self.assertAllClose(m1, m2)

        pass

    def assertMessageToChild(self, X, u):
        self.assertMessage(X._message_to_child(), u)
        pass


    def _get_pack_functions(self, plates, dims):

        inds = np.concatenate(
            [
                [0],
                np.cumsum(
                    [
                        np.prod(dimi) * np.prod(plates)
                        for dimi in dims
                    ]
                )
            ]
        ).astype(int)

        def pack(x):
            return [
                np.reshape(x[start:end], plates + dimi)
                for (start, end, dimi) in zip(inds[:-1], inds[1:], dims)
            ]

        def unpack(u):
            return np.concatenate(
                [
                    np.broadcast_to(ui, plates + dimi).ravel()
                    for (ui, dimi) in zip(u, dims)
                ]
            )

        return (pack, unpack)


    def assert_message_to_parent(self, child, parent, postprocess=lambda u: u,
                                 eps=1e-6, rtol=1e-4, atol=0):

        (pack, unpack) = self._get_pack_functions(parent.plates, parent.dims)

        def cost(x):
            parent.u = pack(x)
            return child.lower_bound_contribution()

        d = postprocess(pack(unpack(parent._message_from_children())))

        d_num = postprocess(
            pack(
                approx_fprime(
                    unpack(parent.u),
                    cost,
                    eps
                )
            )
        )

        # for (i, j) in zip(postprocess(pack(d)), postprocess(pack(d_num))):
        #     print(i)
        #     print(j)

        assert len(d_num) == len(d)

        for i in range(len(d)):
            self.assertAllClose(d[i], d_num[i], rtol=rtol, atol=atol)


    def assert_moments(self, node, postprocess=lambda u: u, eps=1e-6,
                       rtol=1e-4, atol=0):

        (u, g) = node._distribution.compute_moments_and_cgf(node.phi)

        (pack, unpack) = self._get_pack_functions(node.plates, node.dims)

        def cost(x):
            (_, g) = node._distribution.compute_moments_and_cgf(pack(x))
            return -np.sum(g)

        u_num = pack(
            approx_fprime(
                unpack(node.phi),
                cost,
                eps
            )
        )

        assert len(u_num) == len(u)

        up = postprocess(u)
        up_num = postprocess(u_num)

        for i in range(len(up)):
            self.assertAllClose(up[i], up_num[i], rtol=rtol, atol=atol)

        pass



def symm(X):
    """
    Make X symmetric.
    """
    return 0.5 * (X + np.swapaxes(X, -1, -2))

def unique(l):
    """
    Remove duplicate items from a list while preserving order.
    """
    seen = set()
    seen_add = seen.add
    return [ x for x in l if x not in seen and not seen_add(x)]

def tempfile(prefix='', suffix=''):
    return tmp.NamedTemporaryFile(prefix=prefix, suffix=suffix).name

def write_to_hdf5(group, data, name):
    """
    Writes the given array into the HDF5 file.
    """
    try:
        # Try using compression. It doesn't work for scalars.
        group.create_dataset(name,
                             data=data,
                             compression='gzip')
    except TypeError:
        group.create_dataset(name,
                             data=data)
    except ValueError:
        raise ValueError('Could not write %s' % data)


def nans(size=()):
    return np.tile(np.nan, size)

def trues(shape):
    return np.ones(shape, dtype=np.bool)

def identity(*shape):
    return np.reshape(np.identity(np.prod(shape)), shape+shape)

def array_to_scalar(x):
    # This transforms an N-dimensional array to a scalar. It's most
    # useful when you know that the array has only one element and you
    # want it out as a scalar.
    return np.ravel(x)[0]

#def diag(x):


def put(x, indices, y, axis=-1, ufunc=np.add):
    """A kind of inverse mapping of `np.take`

    In a simple, the operation can be thought as:

    .. code-block:: python

       x[indices] += y

    with the exception that all entries of `y` are used instead of just the
    first occurence corresponding to a particular element. That is, the results
    are accumulated, and the accumulation function can be changed by providing
    `ufunc`. For instance, `np.multiply` corresponds to:

    .. code-block:: python

       x[indices] *= y

    Whereas `np.take` picks indices along an axis and returns the resulting
    array, `put` similarly picks indices along an axis but accumulates the
    given values to those entries.

    Example
    -------

    .. code-block:: python

       >>> x = np.zeros(3)
       >>> put(x, [2, 2, 0, 2, 2], 1)
       array([1., 0., 4.])

    `y` must broadcast to the shape of `np.take(x, indices)`:

    .. code-block:: python

       >>> x = np.zeros((3,4))
       >>> put(x, [[2, 2, 0, 2, 2], [1, 2, 1, 2, 1]], np.ones((2,1,4)), axis=0)
       array([[1., 1., 1., 1.],
              [3., 3., 3., 3.],
              [6., 6., 6., 6.]])

    """
    #x = np.copy(x)
    ndim = np.ndim(x)
    if not isinstance(axis, int):
        raise ValueError("Axis must be an integer")

    # Make axis index positive: [0, ..., ndim-1]
    if axis < 0:
        axis = axis + ndim
    if axis < 0 or axis >= ndim:
        raise ValueError("Axis out of bounds")

    indices = axis*(slice(None),) + (indices,) + (ndim-axis-1)*(slice(None),)
    #y = add_trailing_axes(y, ndim-axis-1)
    ufunc.at(x, indices, y)
    return x


def put_simple(y, indices, axis=-1, length=None):
    """An inverse operation of `np.take` with accumulation and broadcasting.

    Compared to `put`, the difference is that the result array is initialized
    with an array of zeros whose shape is determined automatically and `np.add`
    is used as the accumulator.

    """

    if length is None:
        # Try to determine the original length of the axis by finding the
        # largest index. It is more robust to give the length explicitly.
        indices = np.copy(indices)
        indices[indices<0] = np.abs(indices[indices<0]) - 1
        length = np.amax(indices) + 1

    if not isinstance(axis, int):
        raise ValueError("Axis must be an integer")

    # Make axis index negative: [-ndim, ..., -1]
    if axis >= 0:
        raise ValueError("Axis index must be negative")

    y = atleast_nd(y, abs(axis)-1)
    shape_y = np.shape(y)
    end_before = axis - np.ndim(indices) + 1
    start_after = axis + 1
    if end_before == 0:
        shape_x = shape_y + (length,)
    elif start_after == 0:
        shape_x = shape_y[:end_before] + (length,)
    else:
        shape_x = shape_y[:end_before] + (length,) + shape_y[start_after:]

    x = np.zeros(shape_x)

    return put(x, indices, y, axis=axis)


def grid(x1, x2):
    """ Returns meshgrid as a (M*N,2)-shape array. """
    (X1, X2) = np.meshgrid(x1, x2)
    return np.hstack((X1.reshape((-1,1)),X2.reshape((-1,1))))


# class CholeskyDense():

#     def __init__(self, K):
#         self.U = linalg.cho_factor(K)

#     def solve(self, b):
#         if sparse.issparse(b):
#             b = b.toarray()
#         return linalg.cho_solve(self.U, b)

#     def logdet(self):
#         return 2*np.sum(np.log(np.diag(self.U[0])))

#     def trace_solve_gradient(self, dK):
#         return np.trace(self.solve(dK))

# class CholeskySparse():

#     def __init__(self, K):
#         self.LD = cholmod.cholesky(K)

#     def solve(self, b):
#         if sparse.issparse(b):
#             b = b.toarray()
#         return self.LD.solve_A(b)

#     def logdet(self):
#         return self.LD.logdet()
#         #np.sum(np.log(LD.D()))

#     def trace_solve_gradient(self, dK):
#         # WTF?! numpy.multiply doesn't work for two sparse
#         # matrices.. It returns a result but it is incorrect!

#         # Use the identity trace(K\dK)=sum(inv(K).*dK) by computing
#         # the sparse inverse (lower triangular part)
#         iK = self.LD.spinv(form='lower')
#         return (2*iK.multiply(dK).sum()
#                 - iK.diagonal().dot(dK.diagonal()))
#         # Multiply by two because of symmetry (remove diagonal once
#         # because it was taken into account twice)
#         #return np.multiply(self.LD.inv().todense(),dK.todense()).sum()
#         #return self.LD.inv().multiply(dK).sum() # THIS WORKS
#         #return np.multiply(self.LD.inv(),dK).sum() # THIS NOT WORK!! WTF??
#         iK = self.LD.spinv()
#         return iK.multiply(dK).sum()
#         #return (2*iK.multiply(dK).sum()
#         #        - iK.diagonal().dot(dK.diagonal()))
#         #return (2*np.multiply(iK, dK).sum()
#         #        - iK.diagonal().dot(dK.diagonal())) # THIS NOT WORK!!
#         #return np.trace(self.solve(dK))


# def cholesky(K):
#     if isinstance(K, np.ndarray):
#         return CholeskyDense(K)
#     elif sparse.issparse(K):
#         return CholeskySparse(K)
#     else:
#         raise Exception("Unsupported covariance matrix type")

# Computes log probability density function of the Gaussian
# distribution
def gaussian_logpdf(y_invcov_y,
                    y_invcov_mu,
                    mu_invcov_mu,
                    logdetcov,
                    D):

    return (-0.5*D*np.log(2*np.pi)
            -0.5*logdetcov
            -0.5*y_invcov_y
            +y_invcov_mu
            -0.5*mu_invcov_mu)


def zipper_merge(*lists):
    """
    Combines lists by alternating elements from them.

    Combining lists [1,2,3], ['a','b','c'] and [42,666,99] results in
    [1,'a',42,2,'b',666,3,'c',99]

    The lists should have equal length or they are assumed to have the length of
    the shortest list.

    This is known as alternating merge or zipper merge.
    """

    return list(sum(zip(*lists), ()))

def remove_whitespace(s):
    return ''.join(s.split())

def is_numeric(a):
    return (np.isscalar(a) or
            isinstance(a, list) or
            isinstance(a, np.ndarray))

def is_scalar_integer(x):
    t = np.asanyarray(x).dtype.type
    return np.ndim(x) == 0 and issubclass(t, np.integer)


def isinteger(x):
    t = np.asanyarray(x).dtype.type
    return ( issubclass(t, np.integer) or issubclass(t, np.bool_) )


def is_string(s):
    return isinstance(s, str)

def multiply_shapes(*shapes):
    """
    Compute element-wise product of lists/tuples.

    Shorter lists are concatenated with leading 1s in order to get lists with
    the same length.
    """

    # Make the shapes equal length
    shapes = make_equal_length(*shapes)

    # Compute element-wise product
    f = lambda X,Y: (x*y for (x,y) in zip(X,Y))
    shape = functools.reduce(f, shapes)

    return tuple(shape)

def make_equal_length(*shapes):
    """
    Make tuples equal length.

    Add leading 1s to shorter tuples.
    """

    # Get maximum length
    max_len = max((len(shape) for shape in shapes))

    # Make the shapes equal length
    shapes = ((1,)*(max_len-len(shape)) + tuple(shape) for shape in shapes)

    return shapes


def make_equal_ndim(*arrays):
    """
    Add trailing unit axes so that arrays have equal ndim
    """
    shapes = [np.shape(array) for array in arrays]
    shapes = make_equal_length(*shapes)
    arrays = [np.reshape(array, shape)
              for (array, shape) in zip(arrays, shapes)]
    return arrays


def sum_to_dim(A, dim):
    """
    Sum leading axes of A such that A has dim dimensions.
    """
    dimdiff = np.ndim(A) - dim
    if dimdiff > 0:
        axes = np.arange(dimdiff)
        A = np.sum(A, axis=axes)
    return A


def broadcasting_multiplier(plates, *args):
    """
    Compute the plate multiplier for given shapes.

    The first shape is compared to all other shapes (using NumPy
    broadcasting rules). All the elements which are non-unit in the first
    shape but 1 in all other shapes are multiplied together.

    This method is used, for instance, for computing a correction factor for
    messages to parents: If this node has non-unit plates that are unit
    plates in the parent, those plates are summed. However, if the message
    has unit axis for that plate, it should be first broadcasted to the
    plates of this node and then summed to the plates of the parent. In
    order to avoid this broadcasting and summing, it is more efficient to
    just multiply by the correct factor. This method computes that
    factor. The first argument is the full plate shape of this node (with
    respect to the parent). The other arguments are the shape of the message
    array and the plates of the parent (with respect to this node).
    """

    # Check broadcasting of the shapes
    for arg in args:
        broadcasted_shape(plates, arg)

    # Check that each arg-plates are a subset of plates?
    for arg in args:
        if not is_shape_subset(arg, plates):
            print("Plates:", plates)
            print("Args:", args)
            raise ValueError("The shapes in args are not a sub-shape of "
                             "plates")

    r = 1
    for j in range(-len(plates),0):
        mult = True
        for arg in args:
            # if -j <= len(arg) and arg[j] != 1:
            if not (-j > len(arg) or arg[j] == 1):
                mult = False
        if mult:
            r *= plates[j]
    return r


def sum_multiply_to_plates(*arrays, to_plates=(), from_plates=None, ndim=0):
    """
    Compute the product of the arguments and sum to the target shape.
    """
    arrays = list(arrays)
    def get_plates(x):
        if ndim == 0:
            return x
        else:
            return x[:-ndim]

    plates_arrays = [get_plates(np.shape(array)) for array in arrays]
    product_plates = broadcasted_shape(*plates_arrays)

    if from_plates is None:
        from_plates = product_plates
        r = 1
    else:
        r = broadcasting_multiplier(from_plates, product_plates, to_plates)

    for ind in range(len(arrays)):
        plates_others = plates_arrays[:ind] + plates_arrays[(ind+1):]
        plates_without = broadcasted_shape(to_plates, *plates_others)
        ax = axes_to_collapse(plates_arrays[ind], #get_plates(np.shape(arrays[ind])),
                              plates_without)
        if ax:
            ax = tuple([a-ndim for a in ax])
            arrays[ind] = np.sum(arrays[ind], axis=ax, keepdims=True)

    plates_arrays = [get_plates(np.shape(array)) for array in arrays]
    product_plates = broadcasted_shape(*plates_arrays)

    ax = axes_to_collapse(product_plates, to_plates)
    if ax:
        ax = tuple([a-ndim for a in ax])
        y = sum_multiply(*arrays, axis=ax, keepdims=True)
    else:
        y = functools.reduce(np.multiply, arrays)
    y = squeeze_to_dim(y, len(to_plates) + ndim)
    return r * y


def multiply(*arrays):
    return functools.reduce(np.multiply, arrays, 1)


def sum_multiply(*args, axis=None, sumaxis=True, keepdims=False):

    # Computes sum(arg[0]*arg[1]*arg[2]*..., axis=axes_to_sum) without
    # explicitly computing the intermediate product

    if len(args) == 0:
        raise ValueError("You must give at least one input array")

    # Dimensionality of the result
    max_dim = 0
    for k in range(len(args)):
        max_dim = max(max_dim, np.ndim(args[k]))

    if sumaxis:
        if axis is None:
            # Sum all axes
            axes = []
        else:
            if np.isscalar(axis):
                axis = [axis]
            axes = [i
                    for i in range(max_dim)
                    if i not in axis and (-max_dim+i) not in axis]
    else:
        if axis is None:
            # Keep all axes
            axes = list(range(max_dim))
        else:
            # Find axes that are kept
            if np.isscalar(axis):
                axes = [axis]
            axes = [i if i >= 0
                    else i+max_dim
                    for i in axis]
            axes = sorted(axes)

    if len(axes) > 0 and (min(axes) < 0 or max(axes) >= max_dim):
        raise ValueError("Axis index out of bounds")

    # Form a list of pairs: the array in the product and its axes
    pairs = list()
    for i in range(len(args)):
        a = args[i]
        a_dim = np.ndim(a)
        pairs.append(a)
        pairs.append(range(max_dim-a_dim, max_dim))

    # Output axes are those which are not summed
    pairs.append(axes)

    # Compute the sum-product
    try:
        # Set optimize=False to work around a einsum broadcasting bug in NumPy 1.14.0:
        # https://github.com/numpy/numpy/issues/10343
        # Perhaps it'll be fixed in 1.14.1?
        y = np.einsum(*pairs, optimize=False)
    except ValueError as err:
        if str(err) == ("If 'op_axes' or 'itershape' is not NULL in "
                        "theiterator constructor, 'oa_ndim' must be greater "
                        "than zero"):
            # TODO/FIXME: Handle a bug in NumPy. If all arguments to einsum are
            # scalars, it raises an error. For scalars we can just use multiply
            # and forget about summing. Hopefully, in the future, einsum handles
            # scalars properly and this try-except becomes unnecessary.
            y = functools.reduce(np.multiply, args)
        else:
            raise err

    # Restore summed axes as singleton axes
    if keepdims:
        d = 0
        s = ()
        for k in range(max_dim):
            if k in axes:
                # Axis not summed
                s = s + (np.shape(y)[d],)
                d += 1
            else:
                # Axis was summed
                s = s + (1,)
        y = np.reshape(y, s)

    return y

def sum_product(*args, axes_to_keep=None, axes_to_sum=None, keepdims=False):
    if axes_to_keep is not None:
        return sum_multiply(*args,
                            axis=axes_to_keep,
                            sumaxis=False,
                            keepdims=keepdims)
    else:
        return sum_multiply(*args,
                            axis=axes_to_sum,
                            sumaxis=True,
                            keepdims=keepdims)

def moveaxis(A, axis_from, axis_to):
    """
    Move the axis `axis_from` to position `axis_to`.
    """
    if ((axis_from < 0 and abs(axis_from) > np.ndim(A)) or
        (axis_from >= 0 and axis_from >= np.ndim(A)) or
        (axis_to < 0 and abs(axis_to) > np.ndim(A)) or
        (axis_to >= 0  and axis_to >= np.ndim(A))):

        raise ValueError("Can't move axis %d to position %d. Axis index out of "
                         "bounds for array with shape %s"
                         % (axis_from,
                            axis_to,
                            np.shape(A)))

    axes = np.arange(np.ndim(A))
    axes[axis_from:axis_to] += 1
    axes[axis_from:axis_to:-1] -= 1
    axes[axis_to] = axis_from
    return np.transpose(A, axes=axes)


def safe_indices(inds, shape):
    """
    Makes sure that indices are valid for given shape.

    The shorter shape determines the length.

    For instance,

    .. testsetup::

       from bayespy.utils.misc import safe_indices

    >>> safe_indices( (3, 4, 5), (1, 6) )
    (0, 5)
    """
    m = min(len(inds), len(shape))

    if m == 0:
        return ()

    inds = inds[-m:]
    maxinds = np.array(shape[-m:]) - 1

    return tuple(np.fmin(inds, maxinds))


def broadcasted_shape(*shapes):
    """
    Computes the resulting broadcasted shape for a given set of shapes.

    Uses the broadcasting rules of NumPy.  Raises an exception if the shapes do
    not broadcast.
    """
    dim = 0
    for a in shapes:
        dim = max(dim, len(a))
    S = ()
    for i in range(-dim,0):
        s = 1
        for a in shapes:
            if -i <= len(a):
                if s == 1:
                    s = a[i]
                elif a[i] != 1 and a[i] != s:
                    raise ValueError("Shapes %s do not broadcast" % (shapes,))
        S = S + (s,)
    return S

def broadcasted_shape_from_arrays(*arrays):
    """
    Computes the resulting broadcasted shape for a given set of arrays.

    Raises an exception if the shapes do not broadcast.
    """

    shapes = [np.shape(array) for array in arrays]
    return broadcasted_shape(*shapes)


def is_shape_subset(sub_shape, full_shape):
    """
    """
    if len(sub_shape) > len(full_shape):
        return False
    for i in range(len(sub_shape)):
        ind = -1 - i
        if sub_shape[ind] != 1 and sub_shape[ind] != full_shape[ind]:
            return False
    return True


def add_axes(X, num=1, axis=0):
    for i in range(num):
        X = np.expand_dims(X, axis=axis)
    return X
    shape = np.shape(X)[:axis] + num*(1,) + np.shape(X)[axis:]
    return np.reshape(X, shape)


def add_leading_axes(x, n):
    return add_axes(x, axis=0, num=n)


def add_trailing_axes(x, n):
    return add_axes(x, axis=-1, num=n)


def nested_iterator(max_inds):
    s = [range(i) for i in max_inds]
    return itertools.product(*s)

def first(L):
    """
    """
    for (n,l) in enumerate(L):
        if l:
            return n
    return None

def squeeze(X):
    """
    Remove leading axes that have unit length.

    For instance, a shape (1,1,4,1,3) will be reshaped to (4,1,3).
    """
    shape = np.array(np.shape(X))
    inds = np.nonzero(shape != 1)[0]
    if len(inds) == 0:
        shape = ()
    else:
        shape = shape[inds[0]:]
    return np.reshape(X, shape)


def squeeze_to_dim(X, dim):
    s = tuple(range(np.ndim(X)-dim))
    return np.squeeze(X, axis=s)


def axes_to_collapse(shape_x, shape_to):
    # Solves which axes of shape shape_x need to be collapsed in order
    # to get the shape shape_to
    s = ()
    for j in range(-len(shape_x), 0):
        if shape_x[j] != 1:
            if -j > len(shape_to) or shape_to[j] == 1:
                s += (j,)
            elif shape_to[j] != shape_x[j]:
                print('Shape from: ' + str(shape_x))
                print('Shape to: ' + str(shape_to))
                raise Exception('Incompatible shape to squeeze')
    return tuple(s)

def sum_to_shape(X, s):
    """
    Sum axes of the array such that the resulting shape is as given.

    Thus, the shape of the result will be s or an error is raised.
    """
    # First, sum and remove axes that are not in s
    if np.ndim(X) > len(s):
        axes = tuple(range(-np.ndim(X), -len(s)))
    else:
        axes = ()
    Y = np.sum(X, axis=axes)

    # Second, sum axes that are 1 in s but keep the axes
    axes = ()
    for i in range(-np.ndim(Y), 0):
        if s[i] == 1:
            if np.shape(Y)[i] > 1:
                axes = axes + (i,)
        else:
            if np.shape(Y)[i] != s[i]:
                raise ValueError("Shape %s can't be summed to shape %s" %
                                 (np.shape(X), s))
    Y = np.sum(Y, axis=axes, keepdims=True)

    return Y

def repeat_to_shape(A, s):
    # Current shape
    t = np.shape(A)
    if len(t) > len(s):
        raise Exception("Can't repeat to a smaller shape")
    # Add extra axis
    t = tuple([1]*(len(s)-len(t))) + t
    A = np.reshape(A,t)
    # Repeat
    for i in reversed(range(len(s))):
        if s[i] != t[i]:
            if t[i] != 1:
                raise Exception("Can't repeat non-singular dimensions")
            else:
                A = np.repeat(A, s[i], axis=i)
    return A

def multidigamma(a, d):
    """
    Returns the derivative of the log of multivariate gamma.
    """
    return np.sum(special.digamma(a[...,None] - 0.5*np.arange(d)),
                  axis=-1)

m_digamma = multidigamma


def diagonal(A):
    return np.diagonal(A, axis1=-2, axis2=-1)


def make_diag(X, ndim=1, ndim_from=0):
    """
    Create a diagonal array given the diagonal elements.

    The diagonal array can be multi-dimensional. By default, the last axis is
    transformed to two axes (diagonal matrix) but this can be changed using ndim
    keyword. For instance, an array with shape (K,L,M,N) can be transformed to a
    set of diagonal 4-D tensors with shape (K,L,M,N,M,N) by giving ndim=2. If
    ndim=3, the result has shape (K,L,M,N,L,M,N), and so on.

    Diagonality means that for the resulting array Y holds:
    Y[...,i_1,i_2,..,i_ndim,j_1,j_2,..,j_ndim] is zero if i_n!=j_n for any n.
    """
    if ndim < 0:
        raise ValueError("Parameter ndim must be non-negative integer")

    if ndim_from < 0:
        raise ValueError("Parameter ndim_to must be non-negative integer")

    if ndim_from > ndim:
        raise ValueError("Parameter ndim_to must not be greater than ndim")

    if ndim == 0:
        return X

    if np.ndim(X) < 2 * ndim_from:
        raise ValueError("The array does not have enough axes")

    if ndim_from > 0:
        if np.shape(X)[-ndim_from:] != np.shape(X)[-2*ndim_from:-ndim_from]:
            raise ValueError("The array X is not square")

    if ndim == ndim_from:
        return X

    X = atleast_nd(X, ndim+ndim_from)

    if ndim > 0:
        if ndim_from > 0:
            I = identity(*(np.shape(X)[-(ndim_from+ndim):-ndim_from]))
        else:
            I = identity(*(np.shape(X)[-ndim:]))
        X = add_axes(X, axis=np.ndim(X)-ndim_from, num=ndim-ndim_from)
        X = I * X
    return X


def get_diag(X, ndim=1, ndim_to=0):
    """
    Get the diagonal of an array.

    If ndim>1, take the diagonal of the last 2*ndim axes.
    """
    if ndim < 0:
        raise ValueError("Parameter ndim must be non-negative integer")

    if ndim_to < 0:
        raise ValueError("Parameter ndim_to must be non-negative integer")

    if ndim_to > ndim:
        raise ValueError("Parameter ndim_to must not be greater than ndim")

    if ndim == 0:
        return X

    if np.ndim(X) < 2*ndim:
        raise ValueError("The array does not have enough axes")

    if np.shape(X)[-ndim:] != np.shape(X)[-2*ndim:-ndim]:
        raise ValueError("The array X is not square")

    if ndim == ndim_to:
        return X

    n_plate_axes = np.ndim(X) - 2 * ndim
    n_diag_axes = ndim - ndim_to

    axes = tuple(range(0, np.ndim(X) - ndim + ndim_to))

    lengths = [0, n_plate_axes, n_diag_axes, ndim_to, ndim_to]
    cutpoints = list(np.cumsum(lengths))

    axes_plates = axes[cutpoints[0]:cutpoints[1]]
    axes_diag= axes[cutpoints[1]:cutpoints[2]]
    axes_dims1 = axes[cutpoints[2]:cutpoints[3]]
    axes_dims2 = axes[cutpoints[3]:cutpoints[4]]

    axes_input = axes_plates + axes_diag + axes_dims1 + axes_diag + axes_dims2
    axes_output = axes_plates + axes_diag + axes_dims1 + axes_dims2

    return np.einsum(X, axes_input, axes_output)


def diag(X, ndim=1):
    """
    Create a diagonal array given the diagonal elements.

    The diagonal array can be multi-dimensional. By default, the last axis is
    transformed to two axes (diagonal matrix) but this can be changed using ndim
    keyword. For instance, an array with shape (K,L,M,N) can be transformed to a
    set of diagonal 4-D tensors with shape (K,L,M,N,M,N) by giving ndim=2. If
    ndim=3, the result has shape (K,L,M,N,L,M,N), and so on.

    Diagonality means that for the resulting array Y holds:
    Y[...,i_1,i_2,..,i_ndim,j_1,j_2,..,j_ndim] is zero if i_n!=j_n for any n.
    """
    X = atleast_nd(X, ndim)
    if ndim > 0:
        I = identity(*(np.shape(X)[-ndim:]))
        X = add_axes(X, axis=np.ndim(X), num=ndim)
        X = I * X
    return X

def m_dot(A,b):
    # Compute matrix-vector product over the last two axes of A and
    # the last axes of b.  Other axes are broadcasted. If A has shape
    # (..., M, N) and b has shape (..., N), then the result has shape
    # (..., M)

    #b = reshape(b, shape(b)[:-1] + (1,) + shape(b)[-1:])
    #return np.dot(A, b)
    return np.einsum('...ik,...k->...i', A, b)
    # TODO: Use einsum!!
    #return np.sum(A*b[...,np.newaxis,:], axis=(-1,))


def block_banded(D, B):
    """
    Construct a symmetric block-banded matrix.

    `D` contains square diagonal blocks.
    `B` contains super-diagonal blocks.

    The resulting matrix is:

    D[0],   B[0],   0,    0,    ..., 0,        0,        0
    B[0].T, D[1],   B[1], 0,    ..., 0,        0,        0
    0,      B[1].T, D[2], B[2], ..., ...,      ...,      ...
    ...     ...     ...   ...   ..., B[N-2].T, D[N-1],   B[N-1]
    0,      0,      0,    0,    ..., 0,        B[N-1].T, D[N]

    """

    D = [np.atleast_2d(d) for d in D]
    B = [np.atleast_2d(b) for b in B]

    # Number of diagonal blocks
    N = len(D)

    if len(B) != N-1:
        raise ValueError("The number of super-diagonal blocks must contain "
                         "exactly one block less than the number of diagonal "
                         "blocks")

    # Compute the size of the full matrix
    M = 0
    for i in range(N):
        if np.ndim(D[i]) != 2:
            raise ValueError("Blocks must be 2 dimensional arrays")
        d = np.shape(D[i])
        if d[0] != d[1]:
            raise ValueError("Diagonal blocks must be square")
        M += d[0]

    for i in range(N-1):
        if np.ndim(B[i]) != 2:
            raise ValueError("Blocks must be 2 dimensional arrays")
        b = np.shape(B[i])
        if b[0] != np.shape(D[i])[1] or b[1] != np.shape(D[i+1])[0]:
            raise ValueError("Shapes of the super-diagonal blocks do not match "
                             "the shapes of the diagonal blocks")

    A = np.zeros((M,M))
    k = 0

    for i in range(N-1):
        (d0, d1) = np.shape(B[i])
        # Diagonal block
        A[k:k+d0, k:k+d0] = D[i]
        # Super-diagonal block
        A[k:k+d0, k+d0:k+d0+d1] = B[i]
        # Sub-diagonal block
        A[k+d0:k+d0+d1, k:k+d0] = B[i].T

        k += d0
    A[k:,k:] = D[-1]

    return A


def dist_haversine(c1, c2, radius=6372795):

    # Convert coordinates to radians
    lat1 = np.atleast_1d(c1[0])[...,:,None] * np.pi / 180
    lon1 = np.atleast_1d(c1[1])[...,:,None] * np.pi / 180
    lat2 = np.atleast_1d(c2[0])[...,None,:] * np.pi / 180
    lon2 = np.atleast_1d(c2[1])[...,None,:] * np.pi / 180

    dlat = lat2 - lat1
    dlon = lon2 - lon1

    A = np.sin(dlat/2)**2 + np.cos(lat1)*np.cos(lat2)*(np.sin(dlon/2)**2)
    C = 2 * np.arctan2(np.sqrt(A), np.sqrt(1-A))

    return radius * C

def logsumexp(X, axis=None, keepdims=False):
    """
    Compute log(sum(exp(X)) in a numerically stable way
    """

    X = np.asanyarray(X)

    maxX = np.amax(X, axis=axis, keepdims=True)

    if np.ndim(maxX) > 0:
        maxX[~np.isfinite(maxX)] = 0
    elif not np.isfinite(maxX):
        maxX = 0

    X = X - maxX

    if not keepdims:
        maxX = np.squeeze(maxX, axis=axis)

    return np.log(np.sum(np.exp(X), axis=axis, keepdims=keepdims)) + maxX


def normalized_exp(phi):
    """Compute exp(phi) so that exp(phi) sums to one.

    This is useful for computing probabilities from log evidence.
    """
    logsum_p = logsumexp(phi, axis=-1, keepdims=True)
    logp = phi - logsum_p
    p = np.exp(logp)
    # Because of small numerical inaccuracy, normalize the probabilities
    # again for more accurate results
    return (
        p / np.sum(p, axis=-1, keepdims=True),
        logsum_p
    )


def invpsi(x):
    r"""
    Inverse digamma (psi) function.

    The digamma function is the derivative of the log gamma function.
    This calculates the value Y > 0 for a value X such that digamma(Y) = X.

    For the new version, see Appendix C:
    http://research.microsoft.com/en-us/um/people/minka/papers/dirichlet/minka-dirichlet.pdf

    For the previous implementation, see:
    http://www4.ncsu.edu/~pfackler/

    Are there speed/accuracy differences between the methods?
    """
    x = np.asanyarray(x)

    y = np.where(
        x >= -2.22,
        np.exp(x) + 0.5,
        -1/(x - special.psi(1))
    )
    for i in range(5):
        y = y - (special.psi(y) - x) / special.polygamma(1, y)

    return y

    # # Previous implementation. Is it worse? Is there difference?
    # L = 1.0
    # y = np.exp(x)
    # while (L > 1e-10):
    #     y += L*np.sign(x-special.psi(y))
    #     L /= 2
    # # Ad hoc by Jaakko
    # y = np.where(x < -100, -1 / x, y)
    # return y


def invgamma(x):
    r"""
    Inverse gamma function.

    See: http://mathoverflow.net/a/28977
    """
    k = 1.461632
    c = 0.036534
    L = np.log((x+c)/np.sqrt(2*np.pi))
    W = special.lambertw(L/np.exp(1))
    return L/W + 0.5


def mean(X, axis=None, keepdims=False):
    """
    Compute the mean, ignoring NaNs.
    """
    if np.ndim(X) == 0:
        if axis is not None:
            raise ValueError("Axis out of bounds")
        return X
    X = np.asanyarray(X)
    nans = np.isnan(X)
    X = X.copy()
    X[nans] = 0
    m = (np.sum(X, axis=axis, keepdims=keepdims) /
         np.sum(~nans, axis=axis, keepdims=keepdims))
    return m


def gradient(f, x, epsilon=1e-6):
    return optimize.approx_fprime(x, f, epsilon)


def broadcast(*arrays, ignore_axis=None):
    """
    Explicitly broadcast arrays to same shapes.

    It is possible ignore some axes so that the arrays are not broadcasted
    along those axes.
    """

    shapes = [np.shape(array) for array in arrays]

    if ignore_axis is None:
        full_shape = broadcasted_shape(*shapes)

    else:
        try:
            ignore_axis = tuple(ignore_axis)
        except TypeError:
            ignore_axis = (ignore_axis,)

        if len(ignore_axis) != len(set(ignore_axis)):
            raise ValueError("Indices must be unique")

        if any(i >= 0 for i in ignore_axis):
            raise ValueError("Indices must be negative")

        # Put lengths of ignored axes to 1
        cut_shapes = [
            tuple(
                1
                if i in ignore_axis else
                shape[i]
                for i in range(-len(shape), 0)
            )
            for shape in shapes
        ]

        full_shape = broadcasted_shape(*cut_shapes)

    return [np.ones(full_shape) * array for array in arrays]


def block_diag(*arrays):
    """
    Form a block diagonal array from the given arrays.

    Compared to SciPy's block_diag, this utilizes broadcasting and accepts more
    than dimensions in the input arrays.

    """

    arrays = broadcast(*arrays, ignore_axis=(-1, -2))

    plates = np.shape(arrays[0])[:-2]

    M = sum(np.shape(array)[-2] for array in arrays)
    N = sum(np.shape(array)[-1] for array in arrays)

    Y = np.zeros(plates + (M, N))

    i_start = 0
    j_start = 0
    for array in arrays:
        i_end = i_start + np.shape(array)[-2]
        j_end = j_start + np.shape(array)[-1]
        Y[...,i_start:i_end,j_start:j_end] = array
        i_start = i_end
        j_start = j_end

    return Y


def concatenate(*arrays, axis=-1):
    """
    Concatenate arrays along a given axis.

    Compared to NumPy's concatenate, this utilizes broadcasting.
    """

    # numpy.concatenate doesn't do broadcasting, so we need to do it explicitly
    return np.concatenate(
        broadcast(*arrays, ignore_axis=axis),
        axis=axis
    )



################################################################################
# Copyright (C) 2014 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
Unit tests for `multinomial` module.
"""

import numpy as np
import scipy

from bayespy.nodes import (Multinomial,
                           Dirichlet,
                           Mixture)

from bayespy.utils import random

from bayespy.utils.misc import TestCase


class TestMultinomial(TestCase):
    """
    Unit tests for Multinomial node
    """


    def test_init(self):
        """
        Test the creation of multinomial nodes.
        """

        # Some simple initializations
        X = Multinomial(10, [0.1, 0.3, 0.6])
        X = Multinomial(10, Dirichlet([5,4,3]))

        # Check that plates are correct
        X = Multinomial(10, [0.1, 0.3, 0.6], plates=(3,4))
        self.assertEqual(X.plates,
                         (3,4))
        X = Multinomial(10, 0.25*np.ones((2,3,4)))
        self.assertEqual(X.plates,
                         (2,3))
        n = 10 * np.ones((3,4), dtype=np.int64)
        X = Multinomial(n, [0.1, 0.3, 0.6])
        self.assertEqual(X.plates,
                         (3,4))
        X = Multinomial(n, Dirichlet([2,1,9], plates=(3,4)))
        self.assertEqual(X.plates,
                         (3,4))


        # Probabilities not a vector
        self.assertRaises(ValueError,
                          Multinomial,
                          10,
                          0.5)

        # Invalid probability
        self.assertRaises(ValueError,
                          Multinomial,
                          10,
                          [-0.5, 1.5])
        self.assertRaises(ValueError,
                          Multinomial,
                          10,
                          [0.5, 1.5])

        # Invalid number of trials
        self.assertRaises(ValueError,
                          Multinomial,
                          -1,
                          [0.5, 0.5])
        self.assertRaises(ValueError,
                          Multinomial,
                          8.5,
                          [0.5, 0.5])

        # Inconsistent plates
        self.assertRaises(ValueError,
                          Multinomial,
                          10,
                          0.25*np.ones((2,4)),
                          plates=(3,))

        # Explicit plates too small
        self.assertRaises(ValueError,
                          Multinomial,
                          10,
                          0.25*np.ones((2,4)),
                          plates=(1,))

        pass


    def test_moments(self):
        """
        Test the moments of multinomial nodes.
        """

        # Simple test
        X = Multinomial(1, [0.7,0.2,0.1])
        u = X._message_to_child()
        self.assertEqual(len(u), 1)
        self.assertAllClose(u[0],
                            [0.7,0.2,0.1])

        # Test n
        X = Multinomial(10, [0.7,0.2,0.1])
        u = X._message_to_child()
        self.assertAllClose(u[0],
                            [7,2,1])

        # Test plates in p
        n = np.random.randint(1, 10)
        p = np.random.dirichlet([1,1], size=3)
        X = Multinomial(n, p)
        u = X._message_to_child()
        self.assertAllClose(u[0],
                            p*n)

        # Test plates in n
        n = np.random.randint(1, 10, size=(3,))
        p = np.random.dirichlet([1,1,1,1])
        X = Multinomial(n, p)
        u = X._message_to_child()
        self.assertAllClose(u[0],
                            p*n[:,None])

        # Test plates in p and n
        n = np.random.randint(1, 10, size=(4,1))
        p = np.random.dirichlet([1,1], size=3)
        X = Multinomial(n, p)
        u = X._message_to_child()
        self.assertAllClose(u[0],
                            p*n[...,None])

        # Test with Dirichlet prior
        P = Dirichlet([7, 3])
        logp = P._message_to_child()[0]
        p0 = np.exp(logp[0]) / (np.exp(logp[0]) + np.exp(logp[1]))
        p1 = np.exp(logp[1]) / (np.exp(logp[0]) + np.exp(logp[1]))
        X = Multinomial(1, P)
        u = X._message_to_child()
        p = np.array([p0, p1])
        self.assertAllClose(u[0],
                            p)

        # Test with broadcasted plates
        P = Dirichlet([7, 3], plates=(10,))
        X = Multinomial(5, P)
        u = X._message_to_child()
        self.assertAllClose(u[0] * np.ones(X.get_shape(0)),
                            5*p*np.ones((10,1)))

        pass


    def test_lower_bound(self):
        """
        Test lower bound for multinomial node.
        """

        # Test for a bug found in multinomial
        X = Multinomial(10, [0.3, 0.5, 0.2])
        l = X.lower_bound_contribution()
        self.assertAllClose(l, 0.0)

        pass


    def test_mixture(self):
        """
        Test multinomial mixture
        """

        p0 = [0.1, 0.5, 0.2, 0.2]
        p1 = [0.5, 0.1, 0.1, 0.3]
        p2 = [0.3, 0.2, 0.1, 0.4]
        X = Mixture(2, Multinomial, 10, [p0, p1, p2])
        u = X._message_to_child()
        self.assertAllClose(u[0],
                            10*np.array(p2))

        pass

    def test_mixture_with_count_array(self):
        """
        Test multinomial mixture
        """

        p0 = [0.1, 0.5, 0.2, 0.2]
        p1 = [0.5, 0.1, 0.1, 0.3]
        p2 = [0.3, 0.2, 0.1, 0.4]
        counts = [[10], [5], [3]]
        X = Mixture(2, Multinomial, counts, [p0, p1, p2])
        u = X._message_to_child()
        self.assertAllClose(
            u[0],
            np.array(counts)*np.array(p2)
        )

        # Multi-mixture and count array
        # Shape(p) = (2, 1, 3) + (4,)
        p = [
            [[
                [0.1, 0.5, 0.2, 0.2],
                [0.5, 0.1, 0.1, 0.3],
                [0.3, 0.2, 0.1, 0.4],
            ]],
            [[
                [0.3, 0.2, 0.1, 0.4],
                [0.5, 0.1, 0.2, 0.2],
                [0.4, 0.1, 0.2, 0.3],
            ]],
        ]
        # Shape(Z1) = (1, 3) + (2,) -> () + (2,)
        Z1 = 1
        # Shape(Z2) = (1,) + (3,) -> () + (3,)
        Z2 = 2
        # Shape(counts) = (5, 1)
        counts = [[10], [5], [3], [2], [4]]
        # Shape(X) = (5,) + (4,)
        X = Mixture(
            Z1,
            Mixture,
            Z2,
            Multinomial,
            counts,
            p,
            # NOTE: We mix over axes -3 and -1. But as we first mix over the
            # default (-1), then the next mixing happens over -2 (because one
            # axis was already dropped).
            cluster_plate=-2,
        )
        self.assertAllClose(
            X._message_to_child()[0],
            np.array(counts)[:,0,None] * np.array(p)[Z1,:,Z2]
        )

        # Can't have non-singleton axis in counts over the mixed axis
        p0 = [0.1, 0.5, 0.2, 0.2]
        p1 = [0.5, 0.1, 0.1, 0.3]
        p2 = [0.3, 0.2, 0.1, 0.4]
        counts = [10, 5, 3]
        self.assertRaises(
            ValueError,
            Mixture,
            2,
            Multinomial,
            counts,
            [p0, p1, p2],
        )

        return



################################################################################
# Copyright (C) 2015 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
Unit tests for `concatenate` module.
"""

import warnings

import numpy as np

from bayespy.nodes import (Concatenate,
                           GaussianARD,
                           Gamma)

from bayespy.utils import random

from bayespy.utils.misc import TestCase


class TestConcatenate(TestCase):
    """
    Unit tests for Concatenate node.
    """
    

    def test_init(self):
        """
        Test the creation of Concatenate node
        """

        # One parent only
        X = GaussianARD(0, 1, plates=(3,), shape=())
        Y = Concatenate(X)
        self.assertEqual(Y.plates, (3,))
        self.assertEqual(Y.dims, ( (), () ))

        X = GaussianARD(0, 1, plates=(3,), shape=(2,4))
        Y = Concatenate(X)
        self.assertEqual(Y.plates, (3,))
        self.assertEqual(Y.dims, ( (2,4), (2,4,2,4) ))

        # Two parents
        X1 = GaussianARD(0, 1, plates=(2,), shape=())
        X2 = GaussianARD(0, 1, plates=(3,), shape=())
        Y = Concatenate(X1, X2)
        self.assertEqual(Y.plates, (5,))
        self.assertEqual(Y.dims, ( (), () ))

        # Two parents with shapes
        X1 = GaussianARD(0, 1, plates=(2,), shape=(4,6))
        X2 = GaussianARD(0, 1, plates=(3,), shape=(4,6))
        Y = Concatenate(X1, X2)
        self.assertEqual(Y.plates, (5,))
        self.assertEqual(Y.dims, ( (4,6), (4,6,4,6) ))

        # Two parents with non-default axis
        X1 = GaussianARD(0, 1, plates=(2,4), shape=())
        X2 = GaussianARD(0, 1, plates=(3,4), shape=())
        Y = Concatenate(X1, X2, axis=-2)
        self.assertEqual(Y.plates, (5,4))
        self.assertEqual(Y.dims, ( (), () ))

        # Three parents
        X1 = GaussianARD(0, 1, plates=(2,), shape=())
        X2 = GaussianARD(0, 1, plates=(3,), shape=())
        X3 = GaussianARD(0, 1, plates=(4,), shape=())
        Y = Concatenate(X1, X2, X3)
        self.assertEqual(Y.plates, (9,))
        self.assertEqual(Y.dims, ( (), () ))

        # Constant parent
        X1 = [7.2, 3.5]
        X2 = GaussianARD(0, 1, plates=(3,), shape=())
        Y = Concatenate(X1, X2)
        self.assertEqual(Y.plates, (5,))
        self.assertEqual(Y.dims, ( (), () ))

        # Different moments
        X1 = GaussianARD(0, 1, plates=(3,))
        X2 = Gamma(1, 1, plates=(4,))
        self.assertRaises(ValueError,
                          Concatenate,
                          X1,
                          X2)

        # Incompatible shapes
        X1 = GaussianARD(0, 1, plates=(3,), shape=(2,))
        X2 = GaussianARD(0, 1, plates=(2,), shape=())
        self.assertRaises(ValueError,
                          Concatenate,
                          X1,
                          X2)
        
        # Incompatible plates
        X1 = GaussianARD(0, 1, plates=(4,3), shape=())
        X2 = GaussianARD(0, 1, plates=(5,2,), shape=())
        self.assertRaises(ValueError,
                          Concatenate,
                          X1,
                          X2)
        
        pass
        

    def test_message_to_child(self):
        """
        Test the message to child of Concatenate node.
        """

        var = lambda plates, shape: GaussianARD(
            np.random.randn(*(plates + shape)),
            np.random.rand(*(plates + shape)),
            plates=plates,
            shape=shape
        )

        # Two parents without shapes
        X1 = var((2,), ())
        X2 = var((3,), ())
        Y = Concatenate(X1, X2)
        u1 = X1.get_moments()
        u2 = X2.get_moments()
        u = Y.get_moments()
        self.assertAllClose((u[0]*np.ones((5,)))[:2],
                            u1[0]*np.ones((2,)))
        self.assertAllClose((u[1]*np.ones((5,)))[:2],
                            u1[1]*np.ones((2,)))
        self.assertAllClose((u[0]*np.ones((5,)))[2:],
                            u2[0]*np.ones((3,)))
        self.assertAllClose((u[1]*np.ones((5,)))[2:],
                            u2[1]*np.ones((3,)))

        # Two parents with shapes
        X1 = var((2,), (4,))
        X2 = var((3,), (4,))
        Y = Concatenate(X1, X2)
        u1 = X1.get_moments()
        u2 = X2.get_moments()
        u = Y.get_moments()
        self.assertAllClose((u[0]*np.ones((5,4)))[:2],
                            u1[0]*np.ones((2,4)))
        self.assertAllClose((u[1]*np.ones((5,4,4)))[:2],
                            u1[1]*np.ones((2,4,4)))
        self.assertAllClose((u[0]*np.ones((5,4)))[2:],
                            u2[0]*np.ones((3,4)))
        self.assertAllClose((u[1]*np.ones((5,4,4)))[2:],
                            u2[1]*np.ones((3,4,4)))

        # Test with non-constant axis
        X1 = GaussianARD(0, 1, plates=(2,4), shape=())
        X2 = GaussianARD(0, 1, plates=(3,4), shape=())
        Y = Concatenate(X1, X2, axis=-2)
        u1 = X1.get_moments()
        u2 = X2.get_moments()
        u = Y.get_moments()
        self.assertAllClose((u[0]*np.ones((5,4)))[:2],
                            u1[0]*np.ones((2,4)))
        self.assertAllClose((u[1]*np.ones((5,4)))[:2],
                            u1[1]*np.ones((2,4)))
        self.assertAllClose((u[0]*np.ones((5,4)))[2:],
                            u2[0]*np.ones((3,4)))
        self.assertAllClose((u[1]*np.ones((5,4)))[2:],
                            u2[1]*np.ones((3,4)))

        # Test with constant parent
        X1 = np.random.randn(2, 4)
        X2 = GaussianARD(0, 1, plates=(3,), shape=(4,))
        Y = Concatenate(X1, X2)
        u1 = Y.parents[0].get_moments()
        u2 = X2.get_moments()
        u = Y.get_moments()
        self.assertAllClose((u[0]*np.ones((5,4)))[:2],
                            u1[0]*np.ones((2,4)))
        self.assertAllClose((u[1]*np.ones((5,4,4)))[:2],
                            u1[1]*np.ones((2,4,4)))
        self.assertAllClose((u[0]*np.ones((5,4)))[2:],
                            u2[0]*np.ones((3,4)))
        self.assertAllClose((u[1]*np.ones((5,4,4)))[2:],
                            u2[1]*np.ones((3,4,4)))


        pass


    def test_message_to_parent(self):
        """
        Test the message to parents of Concatenate node.
        """

        # Two parents without shapes
        X1 = GaussianARD(0, 1, plates=(2,), shape=())
        X2 = GaussianARD(0, 1, plates=(3,), shape=())
        Z = Concatenate(X1, X2)
        Y = GaussianARD(Z, 1)
        Y.observe(np.random.randn(*Y.get_shape(0)))
        m1 = X1._message_from_children()
        m2 = X2._message_from_children()
        m = Z._message_from_children()
        self.assertAllClose((m[0]*np.ones((5,)))[:2],
                            m1[0]*np.ones((2,)))
        self.assertAllClose((m[1]*np.ones((5,)))[:2],
                            m1[1]*np.ones((2,)))
        self.assertAllClose((m[0]*np.ones((5,)))[2:],
                            m2[0]*np.ones((3,)))
        self.assertAllClose((m[1]*np.ones((5,)))[2:],
                            m2[1]*np.ones((3,)))

        # Two parents with shapes
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", FutureWarning)

            X1 = GaussianARD(0, 1, plates=(2,), shape=(4,6))
            X2 = GaussianARD(0, 1, plates=(3,), shape=(4,6))
            Z = Concatenate(X1, X2)
            Y = GaussianARD(Z, 1)
            Y.observe(np.random.randn(*Y.get_shape(0)))
            m1 = X1._message_from_children()
            m2 = X2._message_from_children()
            m = Z._message_from_children()
            self.assertAllClose((m[0]*np.ones((5,4,6)))[:2],
                                m1[0]*np.ones((2,4,6)))
            self.assertAllClose((m[1]*np.ones((5,4,6,4,6)))[:2],
                                m1[1]*np.ones((2,4,6,4,6)))
            self.assertAllClose((m[0]*np.ones((5,4,6)))[2:],
                                m2[0]*np.ones((3,4,6)))
            self.assertAllClose((m[1]*np.ones((5,4,6,4,6)))[2:],
                                m2[1]*np.ones((3,4,6,4,6)))

            # Two parents with non-default concatenation axis
            X1 = GaussianARD(0, 1, plates=(2,4), shape=())
            X2 = GaussianARD(0, 1, plates=(3,4), shape=())
            Z = Concatenate(X1, X2, axis=-2)
            Y = GaussianARD(Z, 1)
            Y.observe(np.random.randn(*Y.get_shape(0)))
            m1 = X1._message_from_children()
            m2 = X2._message_from_children()
            m = Z._message_from_children()
            self.assertAllClose((m[0]*np.ones((5,4)))[:2],
                                m1[0]*np.ones((2,4)))
            self.assertAllClose((m[1]*np.ones((5,4)))[:2],
                                m1[1]*np.ones((2,4)))
            self.assertAllClose((m[0]*np.ones((5,4)))[2:],
                                m2[0]*np.ones((3,4)))
            self.assertAllClose((m[1]*np.ones((5,4)))[2:],
                                m2[1]*np.ones((3,4)))

            # Constant parent
            X1 = np.random.randn(2,4,6)
            X2 = GaussianARD(0, 1, plates=(3,), shape=(4,6))
            Z = Concatenate(X1, X2)
            Y = GaussianARD(Z, 1)
            Y.observe(np.random.randn(*Y.get_shape(0)))
            m1 = Z._message_to_parent(0)
            m2 = X2._message_from_children()
            m = Z._message_from_children()
            self.assertAllClose((m[0]*np.ones((5,4,6)))[:2],
                                m1[0]*np.ones((2,4,6)))
            self.assertAllClose((m[1]*np.ones((5,4,6,4,6)))[:2],
                                m1[1]*np.ones((2,4,6,4,6)))
            self.assertAllClose((m[0]*np.ones((5,4,6)))[2:],
                                m2[0]*np.ones((3,4,6)))
            self.assertAllClose((m[1]*np.ones((5,4,6,4,6)))[2:],
                                m2[1]*np.ones((3,4,6,4,6)))

        pass


    def test_mask_to_parent(self):
        """
        Test the mask handling in Concatenate node
        """

        pass



################################################################################
# Copyright (C) 2015 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
Unit tests for `wishart` module.
"""

import numpy as np

from scipy import special

from .. import gaussian
from bayespy.nodes import (Gaussian, 
                           Wishart)

from ...vmp import VB

from bayespy.utils import misc
from bayespy.utils import linalg
from bayespy.utils import random

from bayespy.utils.misc import TestCase

def _student_logpdf(y, mu, Cov, nu):
    D = np.shape(y)[-1]
    return (special.gammaln((nu+D)/2)
            - special.gammaln(nu/2)
            - 0.5 * D * np.log(nu)
            - 0.5 * D * np.log(np.pi)
            - 0.5 * np.linalg.slogdet(Cov)[1]
            - 0.5 * (nu+D) * np.log(1+1/nu*np.einsum('...i,...ij,...j->...',
                                                     y-mu,
                                                     np.linalg.inv(Cov),
                                                     y-mu)))


class TestWishart(TestCase):

    def test_lower_bound(self):
        """
        Test the Wishart VB lower bound
        """

        #
        # By having the Wishart node as the only latent node, VB will give exact
        # results, thus the VB lower bound is the true marginal log likelihood.
        # Thus, check that they are equal. The true marginal likelihood is the
        # multivariate Student-t distribution.
        #

        np.random.seed(42)

        D = 3
        n = (D-1) + np.random.uniform(0.1, 0.5)
        V = random.covariance(D)
        Lambda = Wishart(n, V)
        mu = np.random.randn(D)
        Y = Gaussian(mu, Lambda)
        y = np.random.randn(D)
        Y.observe(y)
        Lambda.update()
        L = Y.lower_bound_contribution() + Lambda.lower_bound_contribution()
        mu = mu
        nu = n + 1 - D
        Cov = V / nu
        self.assertAllClose(L,
                            _student_logpdf(y,
                                            mu,
                                            Cov,
                                            nu))

        pass


    def test_moments(self):
        """
        Test the moments of Wishart node
        """

        np.random.seed(42)

        # Test prior moments
        D = 3
        n = (D-1) + np.random.uniform(0.1,2)
        V = random.covariance(D)
        Lambda = Wishart(n, V)
        Lambda.update()
        u = Lambda.get_moments()
        self.assertAllClose(u[0],
                            n*np.linalg.inv(V),
                            msg='Mean incorrect')
        self.assertAllClose(u[1],
                            (np.sum(special.digamma((n - np.arange(D))/2))
                             + D*np.log(2)
                             - np.linalg.slogdet(V)[1]),
                             msg='Log determinant incorrect')

        # Test posterior moments
        D = 3
        n = (D-1) + np.random.uniform(0.1,2)
        V = random.covariance(D)
        Lambda = Wishart(n, V)
        mu = np.random.randn(D)
        Y = Gaussian(mu, Lambda)
        y = np.random.randn(D)
        Y.observe(y)
        Lambda.update()
        u = Lambda.get_moments()
        n = n + 1
        V = V + np.outer(y-mu, y-mu) 
        self.assertAllClose(u[0],
                            n*np.linalg.inv(V),
                            msg='Mean incorrect')
        self.assertAllClose(u[1],
                            (np.sum(special.digamma((n - np.arange(D))/2))
                             + D*np.log(2)
                             - np.linalg.slogdet(V)[1]),
                             msg='Log determinant incorrect')

        

        pass



################################################################################
# Copyright (C) 2013 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
Unit tests for bayespy.utils.linalg module.
"""

import numpy as np

from .. import misc
from .. import linalg

class TestDot(misc.TestCase):

    def test_dot(self):
        """
        Test dot product multiple multi-dimensional arrays.
        """

        # If no arrays, return 0
        self.assertAllClose(linalg.dot(),
                            0)
        # If only one array, return itself
        self.assertAllClose(linalg.dot([[1,2,3],
                                        [4,5,6]]),
                            [[1,2,3],
                             [4,5,6]])
        # Basic test of two arrays: (2,3) * (3,2)
        self.assertAllClose(linalg.dot([[1,2,3],
                                        [4,5,6]],
                                       [[7,8],
                                        [9,1],
                                        [2,3]]),
                            [[31,19],
                             [85,55]])
        # Basic test of four arrays: (2,3) * (3,2) * (2,1) * (1,2)
        self.assertAllClose(linalg.dot([[1,2,3],
                                        [4,5,6]],
                                       [[7,8],
                                        [9,1],
                                        [2,3]],
                                       [[4],
                                        [5]],
                                       [[6,7]]),
                            [[1314,1533],
                             [3690,4305]])

        # Test broadcasting: (2,2,2) * (2,2,2,2)
        self.assertAllClose(linalg.dot([[[1,2],
                                         [3,4]],
                                        [[5,6],
                                         [7,8]]],
                                       [[[[1,2],
                                          [3,4]],
                                         [[5,6],
                                          [7,8]]],
                                        [[[9,1],
                                          [2,3]],
                                         [[4,5],
                                          [6,7]]]]),
                            [[[[  7,  10],
                               [ 15,  22]],

                              [[ 67,  78],
                               [ 91, 106]]],


                             [[[ 13,   7],
                               [ 35,  15]],

                              [[ 56,  67],
                               [ 76,  91]]]])

        # Inconsistent shapes: (2,3) * (2,3)
        self.assertRaises(ValueError,
                          linalg.dot,
                          [[1,2,3],
                           [4,5,6]],
                          [[1,2,3],
                           [4,5,6]])
        # Other axes do not broadcast: (2,2,2) * (3,2,2)
        self.assertRaises(ValueError,
                          linalg.dot,
                          [[[1,2],
                            [3,4]],
                           [[5,6],
                            [7,8]]],
                          [[[1,2],
                            [3,4]],
                           [[5,6],
                            [7,8]],
                           [[9,1],
                            [2,3]]])
        # Do not broadcast matrix axes: (2,1) * (3,2)
        self.assertRaises(ValueError,
                          linalg.dot,
                          [[1],
                           [2]],
                          [[1,2,3],
                           [4,5,6]])
        # Do not accept less than 2-D arrays: (2) * (2,2)
        self.assertRaises(ValueError,
                          linalg.dot,
                          [1,2],
                          [[1,2,3],
                           [4,5,6]])

class TestBandedSolve(misc.TestCase):

    def test_block_banded_solve(self):
        """
        Test the Gaussian elimination algorithm for block-banded matrices.
        """

        #
        # Create a block-banded matrix
        #

        # Number of blocks
        N = 40

        # Random sizes of the blocks
        #D = np.random.randint(5, 10, size=N)
        # Fixed sizes of the blocks
        D = 5*np.ones(N, dtype=np.int64)

        # Some helpful variables to create the covariances
        W = [np.random.randn(D[i], 2*D[i])
             for i in range(N)]

        # The diagonal blocks (covariances)
        A = [np.dot(W[i], W[i].T) for i in range(N)]
        # The superdiagonal blocks (cross-covariances)
        B = [np.dot(W[i][:,-1:], W[i+1][:,:1].T) for i in range(N-1)]

        C = misc.block_banded(A, B)

        # Create the system to be solved: y=C*x
        x_true = np.random.randn(np.sum(D))
        y = np.dot(C, x_true)
        x_true = np.reshape(x_true, (N, -1))
        y = np.reshape(y, (N, -1))

        #
        # Run tests
        #

        # The correct inverse
        invC = np.linalg.inv(C)

        # Inverse from the function that is tested
        (invA, invB, x, ldet) = linalg.block_banded_solve(np.asarray(A),
                                                          np.asarray(B),
                                                          np.asarray(y))

        # Check that you get the correct number of blocks
        self.assertEqual(len(invA), N)
        self.assertEqual(len(invB), N-1)

        # Check each block
        i0 = 0
        for i in range(N-1):
            i1 = i0 + D[i]
            i2 = i1 + D[i+1]
            # Check diagonal block
            self.assertTrue(np.allclose(invA[i], invC[i0:i1, i0:i1]))
            # Check super-diagonal block
            self.assertTrue(np.allclose(invB[i], invC[i0:i1, i1:i2]))
            i0 = i1
        # Check last block
        self.assertTrue(np.allclose(invA[-1], invC[i0:, i0:]))

        # Check the solution of the system
        self.assertTrue(np.allclose(x_true, x))

        # Check the log determinant
        self.assertAlmostEqual(ldet/np.linalg.slogdet(C)[1], 1)




################################################################################
# Copyright (C) 2015 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
Stochastic variational inference on mixture of Gaussians

Stochastic variational inference is a scalable variational Bayesian
learning method which utilizes stochastic gradient.  For details, see
:cite:`Hoffman:2013`.
"""


import numpy as np
import scipy

import matplotlib.pyplot as plt
import bayespy.plot as myplt

from bayespy.utils import misc
from bayespy.utils import random
from bayespy.nodes import Gaussian, Categorical, Mixture, Dirichlet

from bayespy.inference.vmp.vmp import VB
from bayespy.inference.vmp import transformations

import bayespy.plot as bpplt

from bayespy.demos import pca


def run(N=100000, N_batch=50, seed=42, maxiter=100, plot=True):
    """
    Run deterministic annealing demo for 1-D Gaussian mixture.
    """

    if seed is not None:
        np.random.seed(seed)

    # Number of clusters in the model
    K = 20

    # Dimensionality of the data
    D = 5

    # Generate data
    K_true = 10
    spread = 5
    means = spread * np.random.randn(K_true, D)
    z = random.categorical(np.ones(K_true), size=N)
    data = np.empty((N,D))
    for n in range(N):
        data[n] = means[z[n]] + np.random.randn(D)

    #
    # Standard VB-EM algorithm
    #

    # Full model
    mu = Gaussian(np.zeros(D), np.identity(D),
                  plates=(K,),
                  name='means')
    alpha = Dirichlet(np.ones(K),
                      name='class probabilities')
    Z = Categorical(alpha,
                    plates=(N,),
                    name='classes')
    Y = Mixture(Z, Gaussian, mu, np.identity(D),
                name='observations')

    # Break symmetry with random initialization of the means
    mu.initialize_from_random()

    # Put the data in
    Y.observe(data)

    # Run inference
    Q = VB(Y, Z, mu, alpha)
    Q.save(mu)
    Q.update(repeat=maxiter)
    if plot:
        bpplt.pyplot.plot(np.cumsum(Q.cputime), Q.L, 'k-')
    max_cputime = np.sum(Q.cputime[~np.isnan(Q.cputime)])


    #
    # Stochastic variational inference
    #

    # Construct smaller model (size of the mini-batch)
    mu = Gaussian(np.zeros(D), np.identity(D),
                  plates=(K,),
                  name='means')
    alpha = Dirichlet(np.ones(K),
                      name='class probabilities')
    Z = Categorical(alpha,
                    plates=(N_batch,),
                    plates_multiplier=(N/N_batch,),
                    name='classes')
    Y = Mixture(Z, Gaussian, mu, np.identity(D),
                name='observations')

    # Break symmetry with random initialization of the means
    mu.initialize_from_random()

    # Inference engine
    Q = VB(Y, Z, mu, alpha, autosave_filename=Q.autosave_filename)
    Q.load(mu)

    # Because using mini-batches, messages need to be multiplied appropriately
    print("Stochastic variational inference...")
    Q.ignore_bound_checks = True

    maxiter *= int(N/N_batch)
    delay = 1
    forgetting_rate = 0.7
    for n in range(maxiter):

        # Observe a mini-batch
        subset = np.random.choice(N, N_batch)
        Y.observe(data[subset,:])

        # Learn intermediate variables
        Q.update(Z)

        # Set step length
        step = (n + delay) ** (-forgetting_rate)

        # Stochastic gradient for the global variables
        Q.gradient_step(mu, alpha, scale=step)

        if np.sum(Q.cputime[:n]) > max_cputime:
            break
    
    if plot:
        bpplt.pyplot.plot(np.cumsum(Q.cputime), Q.L, 'r:')

        bpplt.pyplot.xlabel('CPU time (in seconds)')
        bpplt.pyplot.ylabel('VB lower bound')
        bpplt.pyplot.legend(['VB-EM', 'Stochastic inference'], loc='lower right')
        bpplt.pyplot.title('VB for Gaussian mixture model')

    return


if __name__ == '__main__':
    import sys, getopt, os
    try:
        opts, args = getopt.getopt(sys.argv[1:],
                                   "",
                                   ["n=",
                                    "batch=",
                                    "seed=",
                                    "maxiter="])
    except getopt.GetoptError:
        print('python stochastic_inference.py <options>')
        print('--n=<INT>        Number of data points')
        print('--batch=<INT>    Mini-batch size')
        print('--maxiter=<INT>  Maximum number of VB iterations')
        print('--seed=<INT>     Seed (integer) for the random number generator')
        sys.exit(2)

    kwargs = {}
    for opt, arg in opts:
        if opt == "--maxiter":
            kwargs["maxiter"] = int(arg)
        elif opt == "--seed":
            kwargs["seed"] = int(arg)
        elif opt in ("--n",):
            kwargs["N"] = int(arg)
        elif opt in ("--batch",):
            kwargs["N_batch"] = int(arg)

    run(**kwargs)

    plt.show()




################################################################################
# Copyright (C) 2013 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
Unit tests for bayespy.utils.misc module.
"""

import unittest

import warnings

import numpy as np
from scipy.special import psi

from numpy import testing

from .. import misc

class TestCeilDiv(misc.TestCase):

    def test_ceildiv(self):
        """
        Test the ceil division
        """

        self.assertEqual(misc.ceildiv(3, 1),
                         3)
        
        self.assertEqual(misc.ceildiv(6, 3),
                         2)
        
        self.assertEqual(misc.ceildiv(7, 3),
                         3)
        
        self.assertEqual(misc.ceildiv(8, 3),
                         3)
        
        self.assertEqual(misc.ceildiv(-6, 3),
                         -2)
        
        self.assertEqual(misc.ceildiv(-7, 3),
                         -2)
        
        self.assertEqual(misc.ceildiv(-8, 3),
                         -2)
        
        self.assertEqual(misc.ceildiv(-9, 3),
                         -3)
        
        self.assertEqual(misc.ceildiv(6, -3),
                         -2)
        
        self.assertEqual(misc.ceildiv(7, -3),
                         -2)
        
        self.assertEqual(misc.ceildiv(8, -3),
                         -2)
        
        self.assertEqual(misc.ceildiv(9, -3),
                         -3)
        

class TestAddAxes(misc.TestCase):

    def test_add_axes(self):
        """
        Test the add_axes method.
        """
        f = lambda X, **kwargs: np.shape(misc.add_axes(X, **kwargs))

        # By default, add one leading axis
        self.assertEqual(f(np.ones((3,))),
                         (1,3))

        # By default, add leading axes
        self.assertEqual(f(np.ones((3,)), num=3),
                         (1,1,1,3))

        # By default, add one axis               
        self.assertEqual(f(np.ones((3,)), axis=1),
                         (3,1))
                         
        # Add axes to the beginning
        self.assertEqual(f(np.ones((2,3,4,)), axis=0, num=3),
                         (1,1,1,2,3,4))
        
        # Add axes to the middle
        self.assertEqual(f(np.ones((2,3,4,)), axis=1, num=3),
                         (2,1,1,1,3,4))

        # Test negative axis index
        self.assertEqual(f(np.ones((2,3,4,)), axis=-4, num=3),
                         (1,1,1,2,3,4))
        self.assertEqual(f(np.ones((2,3,4,)), axis=-1, num=1),
                         (2,3,4,1))
        self.assertEqual(f(np.ones((2,3,4,)), axis=-2, num=3),
                         (2,3,1,1,1,4))
                         
        # Add axes to the end
        self.assertEqual(f(np.ones((2,3,4,)), axis=3, num=3),
                         (2,3,4,1,1,1))


class TestBroadcasting(unittest.TestCase):

    def test_is_shape_subset(self):
        f = misc.is_shape_subset
        self.assertTrue(f( (), () ))
        self.assertTrue(f( (), (3,) ))
        self.assertTrue(f( (1,), (1,) ))
        self.assertTrue(f( (1,), (3,) ))
        self.assertTrue(f( (1,), (4,1) ))
        self.assertTrue(f( (1,), (4,3) ))
        self.assertTrue(f( (1,), (1,3) ))
        self.assertTrue(f( (3,), (1,3) ))
        self.assertTrue(f( (3,), (4,3) ))
        self.assertTrue(f( (5,1,3), (6,5,4,3) ))
        self.assertTrue(f( (5,4,3), (6,5,4,3) ))

        self.assertFalse(f( (1,), () ))
        self.assertFalse(f( (3,), (1,) ))
        self.assertFalse(f( (4,3,), (3,) ))
        self.assertFalse(f( (4,3,), (1,3,) ))
        self.assertFalse(f( (6,1,4,3,), (6,1,1,3,) ))

class TestMultiplyShapes(unittest.TestCase):

    def test_multiply_shapes(self):
        f = lambda *shapes: tuple(misc.multiply_shapes(*shapes))

        # Basic test
        self.assertEqual(f((2,),
                           (3,)),
                         (6,))
        # Test multiple arguments
        self.assertEqual(f((2,),
                           (3,),
                           (4,)),
                         (24,))
        # Test different lengths and multiple arguments
        self.assertEqual(f((  2,3,),
                           (4,5,6,),
                           (    7,)),
                         (4,10,126,))
        # Test empty shapes
        self.assertEqual(f((),
                           ()),
                         ())
        self.assertEqual(f((),
                           (5,)),
                         (5,))

class TestSumMultiply(unittest.TestCase):

    def check_sum_multiply(self, *shapes, **kwargs):

        # The set of arrays
        x = list()
        for (ind, shape) in enumerate(shapes):
            x += [np.random.randn(*shape)]

        # Result from the function
        yh = misc.sum_multiply(*x,
                               **kwargs)

        axis = kwargs.get('axis', None)
        sumaxis = kwargs.get('sumaxis', True)
        keepdims = kwargs.get('keepdims', False)
        
        # Compute the product
        y = 1
        for xi in x:
            y = y * xi

        # Compute the sum
        if sumaxis:
            y = np.sum(y, axis=axis, keepdims=keepdims)
        else:
            axes = np.arange(np.ndim(y))
            # TODO/FIXME: np.delete has a bug that it doesn't accept negative
            # indices. Thus, transform negative axes to positive axes.
            if len(axis) > 0:
                axis = [i if i >= 0 
                        else i+np.ndim(y) 
                        for i in axis]
            elif axis < 0:
                axis += np.ndim(y)
                
            axes = np.delete(axes, axis)
            axes = tuple(axes)
            if len(axes) > 0:
                y = np.sum(y, axis=axes, keepdims=keepdims)

        # Check the result
        testing.assert_allclose(yh, y,
                                err_msg="Incorrect value.")
        

    def test_sum_multiply(self):
        """
        Test misc.sum_multiply.
        """
        # Check empty list returns error
        self.assertRaises(ValueError, 
                          self.check_sum_multiply)
        # Check scalars
        self.check_sum_multiply(())
        self.check_sum_multiply((), (), ())
        
        # Check doing no summation
        self.check_sum_multiply((3,), 
                                axis=())
        self.check_sum_multiply((3,1,5), 
                                (  4,1),
                                (    5,),
                                (      ),
                                axis=(), 
                                keepdims=True)
        # Check AXES_TO_SUM
        self.check_sum_multiply((3,1), 
                                (1,4),
                                (3,4),
                                axis=(1,))
        self.check_sum_multiply((3,1),
                                (1,4),
                                (3,4),
                                axis=(-2,))
        self.check_sum_multiply((3,1), 
                                (1,4),
                                (3,4),
                                axis=(1,-2))
        # Check AXES_TO_SUM and KEEPDIMS
        self.check_sum_multiply((3,1), 
                                (1,4),
                                (3,4),
                                axis=(1,),
                                keepdims=True)
        self.check_sum_multiply((3,1),
                                (1,4),
                                (3,4),
                                axis=(-2,),
                                keepdims=True)
        self.check_sum_multiply((3,1), 
                                (1,4),
                                (3,4),
                                axis=(1,-2,),
                                keepdims=True)
        self.check_sum_multiply((3,1,5,6), 
                                (  4,1,6),
                                (  4,1,1),
                                (       ),
                                axis=(1,-2),
                                keepdims=True)
        # Check AXES_TO_KEEP
        self.check_sum_multiply((3,1), 
                                (1,4),
                                (3,4),
                                sumaxis=False,
                                axis=(1,))
        self.check_sum_multiply((3,1),
                                (1,4),
                                (3,4),
                                sumaxis=False,
                                axis=(-2,))
        self.check_sum_multiply((3,1), 
                                (1,4),
                                (3,4),
                                sumaxis=False,
                                axis=(1,-2))
        # Check AXES_TO_KEEP and KEEPDIMS
        self.check_sum_multiply((3,1), 
                                (1,4),
                                (3,4),
                                sumaxis=False,
                                axis=(1,),
                                keepdims=True)
        self.check_sum_multiply((3,1),
                                (1,4),
                                (3,4),
                                sumaxis=False,
                                axis=(-2,),
                                keepdims=True)
        self.check_sum_multiply((3,1), 
                                (1,4),
                                (3,4),
                                sumaxis=False,
                                axis=(1,-2,),
                                keepdims=True)
        self.check_sum_multiply((3,1,5,6), 
                                (  4,1,6),
                                (  4,1,1),
                                (       ),
                                sumaxis=False,
                                axis=(1,-2,),
                                keepdims=True)
        # Check errors
        # Inconsistent shapes
        self.assertRaises(ValueError, 
                          self.check_sum_multiply,
                          (3,4),
                          (3,5))
        # Axis index out of bounds
        self.assertRaises(ValueError, 
                          self.check_sum_multiply,
                          (3,4),
                          (3,4),
                          axis=(-3,))
        self.assertRaises(ValueError, 
                          self.check_sum_multiply,
                          (3,4),
                          (3,4),
                          axis=(2,))
        self.assertRaises(ValueError, 
                          self.check_sum_multiply,
                          (3,4),
                          (3,4),
                          sumaxis=False,
                          axis=(-3,))
        self.assertRaises(ValueError, 
                          self.check_sum_multiply,
                          (3,4),
                          (3,4),
                          sumaxis=False,
                          axis=(2,))
        # Same axis several times
        self.assertRaises(ValueError, 
                          self.check_sum_multiply,
                          (3,4),
                          (3,4),
                          axis=(1,-1))
        self.assertRaises(ValueError, 
                          self.check_sum_multiply,
                          (3,4),
                          (3,4),
                          sumaxis=False,
                          axis=(1,-1))

class TestLogSumExp(misc.TestCase):

    def test_logsumexp(self):
        """
        Test the ceil division
        """

        self.assertAllClose(misc.logsumexp(3),
                            np.log(np.sum(np.exp(3))))

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            self.assertAllClose(misc.logsumexp(-np.inf),
                                -np.inf)

        self.assertAllClose(misc.logsumexp(np.inf),
                            np.inf)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            self.assertAllClose(misc.logsumexp(np.nan),
                                np.nan)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            self.assertAllClose(misc.logsumexp([-np.inf, -np.inf]),
                                -np.inf)

        self.assertAllClose(misc.logsumexp([[1e10,  1e-10],
                                             [-1e10, -np.inf]], axis=-1),
                            [1e10, -1e10])

        # Test keeping dimensions
        self.assertAllClose(misc.logsumexp([[1e10,  1e-10],
                                             [-1e10, -np.inf]], 
                                            axis=-1,
                                            keepdims=True),
                            [[1e10], [-1e10]])

        # Test multiple axes
        self.assertAllClose(misc.logsumexp([[1e10,  1e-10],
                                             [-1e10, -np.inf]], 
                                            axis=(-1,-2)),
                            1e10)

        pass

class TestMean(misc.TestCase):

    def test_mean(self):
        """
        Test the ceil division
        """

        self.assertAllClose(misc.mean(3),
                            3)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            self.assertAllClose(misc.mean(np.nan),
                                np.nan)

            self.assertAllClose(misc.mean([[2,3],
                                           [np.nan,np.nan]], 
                                           axis=-1),
                                [2.5,np.nan])
            self.assertAllClose(misc.mean([[2,3],
                                           [np.nan,np.nan]], 
                                           axis=-1,
                                           keepdims=True),
                                [[2.5],[np.nan]])
        
        self.assertAllClose(misc.mean([[2,3],
                                       [np.nan,np.nan]], 
                                       axis=-2),
                            [2,3])
        self.assertAllClose(misc.mean([[2,3],
                                       [np.nan,np.nan]], 
                                       axis=-2,
                                       keepdims=True),
                            [[2,3]])

        self.assertAllClose(misc.mean([[2,3],
                                       [np.nan,np.nan]]),
                            2.5)
        self.assertAllClose(misc.mean([[2,3],
                                       [np.nan,np.nan]],
                                       axis=(-1,-2)),
                            2.5)
        self.assertAllClose(misc.mean([[2,3],
                                       [np.nan,np.nan]], 
                                       keepdims=True),
                            [[2.5]])
        
        pass


class TestInvPsi(misc.TestCase):

    def test_invpsi(self):
        x = 1000
        y = psi(x)
        self.assertAllClose(misc.invpsi(y), x)

        x = 1/1000
        y = psi(x)
        self.assertAllClose(misc.invpsi(y), x, rtol=1e-3)

        x = 50*np.random.rand(5)
        y = psi(x)
        self.assertAllClose(misc.invpsi(y), x)

        pass


class TestPutSimple(misc.TestCase):

    def test_put_simple(self):

        # Scalar indices
        self.assertAllClose(
            misc.put_simple(
                42,
                2,
            ),
            [0, 0, 42],
        )

        # Simple vectors, automatic length
        self.assertAllClose(
            misc.put_simple(
                [1, 0.1, 0.01, 0.001, 0.0001],
                [3, 3, 1, 3, 0],
            ),
            [0.0001, 0.01, 0, 1.101],
        )

        # Matrix indices
        self.assertAllClose(
            misc.put_simple(
                [[1, 0.1], [0.01, 0.001]],
                [[4, 1], [1, 3]],
            ),
            [0, 0.11, 0, 0.001, 1],
        )

        # Test axis
        self.assertAllClose(
            misc.put_simple(
                [[1, 0.1], [0.01, 0.001], [0.0001, 0.00001]],
                [3, 3, 0],
                axis=-2,
            ),
            [[0.0001, 0.00001], [0, 0], [0, 0], [1.01, 0.101]],
        )

        # Test explicit length
        self.assertAllClose(
            misc.put_simple(
                [1, 0.1, 0.01, 0.001, 0.0001],
                [3, 3, 1, 3, 0],
                length=6,
            ),
            [0.0001, 0.01, 0, 1.101, 0, 0],
        )

        # Test broadcasting
        self.assertAllClose(
            misc.put_simple(
                2,
                [3, 3, 1, 3, 0],
            ),
            [2, 2, 0, 6],
        )

        # Test leading axes in y
        self.assertAllClose(
            misc.put_simple(
                [[1, 0.1], [0.01, 0.001], [0.0001, 0.00001]],
                [2, 0],
            ),
            [[0.1, 0, 1], [0.001, 0, 0.01], [0.00001, 0, 0.0001]],
        )

        pass



################################################################################
# Copyright (C) 2011-2014 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


import numpy as np

from bayespy.utils import misc

from .node import Node
from .deterministic import Deterministic
from .gaussian import Gaussian, GaussianMoments
from .gaussian import GaussianGammaMoments
from .ml import DeltaMoments


class SumMultiply(Deterministic):
    r"""
    Node for computing general products and sums of Gaussian nodes.

    The node is similar to `numpy.einsum`, which is a very general
    function for computing dot products, sums, products and other sums
    of products of arrays.

    For instance, consider the following arrays:

    >>> import numpy as np
    >>> X = np.random.randn(2, 3, 4)
    >>> Y = np.random.randn(3, 5)
    >>> Z = np.random.randn(4, 2)

    Then, the Einstein summation can be used as:

    >>> np.einsum('abc,bd,ca->da', X, Y, Z)
    array([[...]])

    SumMultiply node can be used similarly for Gaussian nodes.  For instance,
    consider the following Gaussian nodes:

    >>> from bayespy.nodes import GaussianARD
    >>> X = GaussianARD(0, 1, shape=(2, 3, 4))
    >>> Y = GaussianARD(0, 1, shape=(3, 5))
    >>> Z = GaussianARD(0, 1, shape=(4, 2))

    Then, similarly to `numpy.einsum`, SumMultiply could be used as:

    >>> from bayespy.nodes import SumMultiply
    >>> SumMultiply('abc,bd,ca->da', X, Y, Z)
    <bayespy.inference.vmp.nodes.dot.SumMultiply object at 0x...>

    or

    >>> SumMultiply(X, [0,1,2], Y, [1,3], Z, [2,0], [3,0])
    <bayespy.inference.vmp.nodes.dot.SumMultiply object at 0x...>

    which is similar to the alternative syntax of numpy.einsum.

    This node operates similarly as numpy.einsum. However, you must use all the
    elements of each node, that is, an operation like np.einsum('ii->i',X) is
    not allowed. Thus, for each node, each axis must be given unique id. The id
    identifies which axes correspond to which axes between the different
    nodes. Also, Ellipsis ('...') is not yet supported for simplicity. It would
    also have some problems with constant inputs (because how to determine
    ndim), so let us just forget it for now.

    Each output axis must appear in the input mappings.

    The keys must refer to variable dimension axes only, not plate axes.

    The input nodes may be Gaussian-gamma (isotropic) nodes.

    The output message is Gaussian-gamma (isotropic) if any of the input nodes
    is Gaussian-gamma.

    Examples
    --------

    Sum over the rows:
    'ij->j'

    Inner product of three vectors:
    'i,i,i'

    Matrix-vector product:
    'ij,j->i'

    Matrix-matrix product:
    'ik,kj->ij'

    Outer product:
    'i,j->ij'

    Vector-matrix-vector product:
    'i,ij,j'

    Notes
    -----

    This operation can be extremely slow if not used wisely. For large and
    complex operations, it is sometimes more efficient to split the operation
    into multiple nodes. For instance, the example above could probably be
    computed faster by

    >>> XZ = SumMultiply(X, [0,1,2], Z, [2,0], [0,1])
    >>> F = SumMultiply(XZ, [0,1], Y, [1,2], [2,0])

    because the third axis ('c') could be summed out already in the first
    operation. This same effect applies also to numpy.einsum in general.
    """

    def __init__(self, *args, iterator_axis=None, **kwargs):
        """
        SumMultiply(Node1, map1, Node2, map2, ..., NodeN, mapN [, map_out])
        """

        args = list(args)

        if len(args) < 2:
            raise ValueError("Not enough inputs")

        if iterator_axis is not None:
            raise NotImplementedError("Iterator axis not implemented yet")
        if iterator_axis is not None and not isinstance(iterator_axis, int):
            raise ValueError("Iterator axis must be integer")

        # Two different parsing methods, depends on how the arguments are given
        if misc.is_string(args[0]):
            # This is the format:
            # SumMultiply('ik,k,kj->ij', X, Y, Z)
            strings = args[0]
            nodes = args[1:]
            # Remove whitespace
            strings = misc.remove_whitespace(strings)
            # Split on '->' (should contain only one '->' or none)
            strings = strings.split('->')
            if len(strings) > 2:
                raise ValueError('The string contains too many ->')
            strings_in = strings[0]
            if len(strings) == 2:
                string_out = strings[1]
            else:
                string_out = ''
            # Split former part on ',' (the number of parts should be equal to
            # nodes)
            strings_in = strings_in.split(',')
            if len(strings_in) != len(nodes):
                raise ValueError('Number of given input nodes is different '
                                 'from the input keys in the string')
            # Split strings into key lists using single character keys
            keysets = [list(string_in) for string_in in strings_in]
            keys_out = list(string_out)


        else:
            # This is the format:
            # SumMultiply(X, [0,2], Y, [2], Z, [2,1], [0,1])

            # If given, the output mapping is the last argument
            if len(args) % 2 == 0:
                keys_out = []
            else:
                keys_out = args.pop(-1)

            # Node and axis mapping are given in turns
            nodes = args[::2]
            keysets = args[1::2]

        # Find all the keys (store only once each)
        full_keyset = []
        for keyset in keysets:
            full_keyset += keyset
            #full_keyset += list(keyset.keys())
        full_keyset = list(set(full_keyset))

        # Input and output messages are Gaussian unless there is at least one
        # Gaussian-gamma message from the parents
        self.gaussian_gamma = False
        self.is_constant = []
        for i in range(len(nodes)):
            try:
                # First, try to handle the node as a "constant" with so-called
                # delta moments. These consume much less memory as only the
                # "value" (or the first moment) is used.
                nodes[i] = self._ensure_moments(
                    nodes[i],
                    DeltaMoments,
                    ndim=len(keysets[i])
                )
            except DeltaMoments.NoConverterError:
                self.is_constant.append(False)
                try:
                    # Second, try to convert to gaussian moments. These moments
                    # will consume more memory as it calculates the second
                    # moment too.
                    nodes[i] = self._ensure_moments(
                        nodes[i],
                        GaussianMoments,
                        ndim=len(keysets[i])
                    )
                except GaussianMoments.NoConverterError:
                    self.gaussian_gamma = True
            else:
                self.is_constant.append(True)

        if self.gaussian_gamma:
            nodes = [
                (
                    self._ensure_moments(
                        node,
                        GaussianGammaMoments,
                        ndim=len(keyset)
                    )
                    if not is_const else
                    node
                )
                for (node, keyset, is_const) in zip(nodes, keysets, self.is_constant)
            ]

        self._parent_moments = tuple(node._moments for node in nodes)

        #
        # Check the validity of each node
        #
        for n in range(len(nodes)):
            # Check that the maps and the size of the variable are consistent
            if len(nodes[n].dims[0]) != len(keysets[n]):
                raise ValueError("Wrong number of keys (%d) for the node "
                                 "number %d with %d dimensions"
                                 % (len(keysets[n]),
                                    n,
                                    len(nodes[n].dims[0])))
            # Check that the keys are unique
            if len(set(keysets[n])) != len(keysets[n]):
                raise ValueError("Axis keys for node number %d are not unique"
                                 % n)

        # Check the validity of output keys: each output key must be included in
        # the input keys
        if len(keys_out) != len(set(keys_out)):
            raise ValueError("Output keys are not unique")
        for key in keys_out:
            if key not in full_keyset:
                raise ValueError("Output key %s does not appear in any input"
                                 % key)

        # Check the validity of the nodes with respect to the key mapping.
        # Check that the node dimensions map and broadcast properly, that is,
        # all the nodes using the same key for axes must have equal size for
        # those axes (or size 1).
        broadcasted_size = {}
        for key in full_keyset:
            broadcasted_size[key] = 1
            for (node, keyset) in zip(nodes, keysets):
                try:
                    # Find the axis for the key
                    index = keyset.index(key)
                except ValueError:
                    # OK, this node doesn't use this key for any axis
                    pass
                else:
                    # Length of the axis for that key
                    node_size = node.dims[0][index]
                    if node_size != broadcasted_size[key]:
                        if broadcasted_size[key] == 1:
                            # Apply broadcasting
                            broadcasted_size[key] = node_size
                        elif node_size != 1:
                            # Different sizes and neither has size 1
                            raise ValueError("Axes using key %s do not "
                                             "broadcast properly"
                                             % key)

        # Compute the shape of the output
        shape = tuple([broadcasted_size[key] for key in keys_out])

        if self.gaussian_gamma:
            self._moments = GaussianGammaMoments(shape)
        else:
            self._moments = GaussianMoments(shape)

        # Rename the keys to [0,1,...,N-1] where N is the total number of keys
        self.N_keys = len(full_keyset)
        self.out_keys = [full_keyset.index(key) for key in keys_out]
        self.in_keys = [ [full_keyset.index(key) for key in keyset]
                         for keyset in keysets ]

        super().__init__(*nodes,
                         dims=self._moments.dims,
                         **kwargs)


    def _compute_function(self, *x_parents):

        # TODO: Add unit tests for this function

        (xs, alphas) = (
            (x_parents, 1) if not self.gaussian_gamma else
            zip(*x_parents)
        )

        # Add Ellipsis for the plates
        in_keys = [[Ellipsis] + k for k in self.in_keys]
        out_keys = [Ellipsis] + self.out_keys

        samples_and_keys = misc.zipper_merge(xs, in_keys)
        y = np.einsum(*(samples_and_keys + [out_keys]))

        return (
            y if not self.gaussian_gamma else
            (y, misc.multiply(*alphas))
        )


    def _compute_moments(self, *u_parents):


        # Compute the number of plate axes for each node
        plate_counts0 = [
            (np.ndim(u_parent[0]) - len(keys))
            for (keys,u_parent) in zip(self.in_keys, u_parents)
        ]
        plate_counts1 = [
            (
                # Gaussian moments: Use second moments "matrix"
                (np.ndim(u_parent[1]) - 2*len(keys))
                if not is_const else
                # Delta moments: Use first moment "vector"
                (np.ndim(u_parent[0]) - len(keys))
            )
            for (keys, u_parent, is_const) in zip(
                    self.in_keys,
                    u_parents,
                    self.is_constant
            )
        ]
        # The number of plate axes for the output
        N0 = max(plate_counts0)
        N1 = max(plate_counts1)
        # The total number of unique keys used (keys are 0,1,...,N_keys-1)
        D = self.N_keys

        #
        # Compute the mean
        #
        out_all_keys = list(range(D+N0-1, D-1, -1)) + self.out_keys
        #nodes_dim_keys = self.nodes_dim_keys
        in_all_keys = [list(range(D+plate_count-1, D-1, -1)) + keys
                      for (plate_count, keys) in zip(plate_counts0,
                                                     self.in_keys)]
        u0 = [u[0] for u in u_parents]

        args = misc.zipper_merge(u0, in_all_keys) + [out_all_keys]
        x0 = np.einsum(*args)

        #
        # Compute the covariance
        #
        out_all_keys = (list(range(2*D+N1-1, 2*D-1, -1))
                        + [D+key for key in self.out_keys]
                        + self.out_keys)
        in_all_keys = [
            x
            for (plate_count, node_keys, is_const) in zip(
                    plate_counts1,
                    self.in_keys,
                    self.is_constant,
            )
            for x in (
                    # Gaussian moments: Use the second moment
                    [
                        list(range(2*D+plate_count-1, 2*D-1, -1))
                        + [D+key for key in node_keys]
                        + node_keys
                    ]
                    if not is_const else
                    # Delta moments: Use the first moment tiwce
                    [
                        (
                            list(range(2*D+plate_count-1, 2*D-1, -1))
                            + [D+key for key in node_keys]
                        ),
                        (
                            list(range(2*D+plate_count-1, 2*D-1, -1))
                            + node_keys
                        ),
                    ]
            )
        ]
        u1 = [
            x
            for (u, is_const) in zip(u_parents, self.is_constant)
            for x in (
                    # Gaussian moments: Use the second moment
                    [u[1]]
                    if not is_const else
                    # Delta moments: Use the first moment twice
                    [u[0], u[0]]
            )
        ]
        args = misc.zipper_merge(u1, in_all_keys) + [out_all_keys]
        x1 = np.einsum(*args)

        if not self.gaussian_gamma:
            return [x0, x1]

        # Compute Gaussian-gamma specific moments
        x2 = 1
        x3 = 0
        for i in range(len(u_parents)):
            x2 = x2 * (1 if self.is_constant[i] else u_parents[i][2])
            x3 = x3 + (0 if self.is_constant[i] else u_parents[i][3])

        return [x0, x1, x2, x3]


    def get_parameters(self):
        # Compute mean and variance
        u = self.get_moments()
        u[1] -= u[0]**2
        return u


    def _message_to_parent(self, index, u_parent=None):
        """
        Compute the message and mask to a parent node.
        """

        if self.is_constant[index]:
            raise NotImplementedError(
                "Message to DeltaMoments parent not yet implemented."
            )

        # Check index
        if index >= len(self.parents):
            raise ValueError("Parent index larger than the number of parents")

        # Get messages from other parents and children
        u_parents = self._message_from_parents(exclude=index)
        m = self._message_from_children()
        mask = self.mask

        # Normally we don't need to care about masks when computing the
        # message. However, in this node we want to avoid computing huge message
        # arrays so we sum some axes already here. Thus, we need to apply the
        # mask.
        #
        # Actually, we don't need to care about masks because the message from
        # children has already been masked.

        parent = self.parents[index]

        #
        # Compute the first message
        #

        msg = [None, None]

        # Compute the two messages
        for ind in range(2):

            # The total number of keys for the non-plate dimensions
            N = (ind+1) * self.N_keys

            parent_num_dims = len(parent.dims[ind])
            parent_num_plates = len(parent.plates)
            parent_plate_keys = list(range(N + parent_num_plates,
                                           N,
                                           -1))
            parent_dim_keys = self.in_keys[index]
            if ind == 1:
                parent_dim_keys = ([key + self.N_keys
                                    for key in self.in_keys[index]]
                                   + parent_dim_keys)
            args = []

            # This variable counts the maximum number of plates of the
            # arguments, thus it will tell the number of plates in the result
            # (if the artificially added plates above were ignored).
            result_num_plates = 0
            result_plates = ()

            # Mask and its keysr
            mask_num_plates = np.ndim(mask)
            mask_plates = np.shape(mask)
            mask_plate_keys = list(range(N + mask_num_plates,
                                         N,
                                         -1))
            result_num_plates = max(result_num_plates,
                                    mask_num_plates)
            result_plates = misc.broadcasted_shape(result_plates,
                                                   mask_plates)

            # Moments and keys of other parents
            for (k, u) in enumerate(u_parents):
                if k != index:
                    num_dims = (
                        (ind+1) * len(self.in_keys[k])
                        if not self.is_constant[k] else
                        len(self.in_keys[k])
                    )
                    ui = (
                        u[ind] if not self.is_constant[k] else
                        u[0]
                    )
                    num_plates = np.ndim(ui) - num_dims
                    plates = np.shape(ui)[:num_plates]
                    plate_keys = list(range(N + num_plates,
                                            N,
                                            -1))
                    if ind == 0:
                        args.append(ui)
                        args.append(plate_keys + self.in_keys[k])
                    else:
                        in_keys2 = [key + self.N_keys for key in self.in_keys[k]]
                        if not self.is_constant[k]:
                            # Gaussian moments: Use second moment once
                            args.append(ui)
                            args.append(plate_keys + in_keys2 + self.in_keys[k])
                        else:
                            # Delta moments: Use first moment twice
                            args.append(ui)
                            args.append(plate_keys + self.in_keys[k])
                            args.append(ui)
                            args.append(plate_keys + in_keys2)

                    result_num_plates = max(result_num_plates, num_plates)
                    result_plates = misc.broadcasted_shape(result_plates,
                                                           plates)

            # Message and keys from children
            child_num_dims = (ind+1) * len(self.out_keys)
            child_num_plates = np.ndim(m[ind]) - child_num_dims
            child_plates = np.shape(m[ind])[:child_num_plates]
            child_plate_keys = list(range(N + child_num_plates,
                                          N,
                                          -1))
            child_dim_keys = self.out_keys
            if ind == 1:
                child_dim_keys = ([key + self.N_keys
                                   for key in self.out_keys]
                                  + child_dim_keys)
            args.append(m[ind])
            args.append(child_plate_keys + child_dim_keys)

            result_num_plates = max(result_num_plates, child_num_plates)
            result_plates = misc.broadcasted_shape(result_plates,
                                                   child_plates)

            # Output keys, that is, the keys of the parent[index]
            parent_keys = parent_plate_keys + parent_dim_keys

            # Performance trick: Check which axes can be summed because they
            # have length 1 or are non-existing in parent[index]. Thus, remove
            # keys corresponding to unit length axes in parent[index] so that
            # einsum sums over those axes. After computations, these axes must
            # be added back in order to get the correct shape for the message.
            # Also, remove axes/keys that are in output (parent[index]) but not in
            # any inputs (children and other parents).

            parent_shape = parent.get_shape(ind)
            removed_axes = []
            for j in range(len(parent_keys)):
                if parent_shape[j] == 1:
                    # Remove the key (take into account the number of keys that
                    # have already been removed)
                    del parent_keys[j-len(removed_axes)]
                    removed_axes.append(j)
                else:
                    # Remove the key if it doesn't appear in any of the
                    # messages from children or other parents.
                    if not np.any([parent_keys[j-len(removed_axes)] in keys
                                   for keys in args[1::2]]):
                        del parent_keys[j-len(removed_axes)]
                        removed_axes.append(j)

            args.append(parent_keys)

            # THE BEEF: Compute the message
            msg[ind] = np.einsum(*args)

            # Find the correct shape for the message array
            message_shape = list(np.shape(msg[ind]))
            # First, add back the axes with length 1
            for ax in removed_axes:
                message_shape.insert(ax, 1)
            # Second, remove leading axes for plates that were not present in
            # the child nor other parents' messages. This is not really
            # necessary, but it is just elegant to remove the leading unit
            # length axes that we added artificially at the beginning just
            # because we wanted the key mapping to be simple.
            if parent_num_plates > result_num_plates:
                del message_shape[:(parent_num_plates-result_num_plates)]
            # Then, the actual reshaping
            msg[ind] = np.reshape(msg[ind], message_shape)

            # Broadcasting is not supported for variable dimensions, thus force
            # explicit correct shape for variable dimensions
            var_dims = parent.dims[ind]
            msg[ind] = msg[ind] * np.ones(var_dims)

            # Apply plate multiplier: If this node has non-unit plates that are
            # unit plates in the parent, those plates are summed. However, if
            # the message has unit axis for that plate, it should be first
            # broadcasted to the plates of this node and then summed to the
            # plates of the parent. In order to avoid this broadcasting and
            # summing, it is more efficient to just multiply by the correct
            # factor.
            r = self.broadcasting_multiplier(self.plates,
                                             result_plates,
                                             parent.plates)
            if r != 1:
                msg[ind] *= r

        if self.gaussian_gamma:
            alphas = [
                (u_parents[i][2] if not is_const else 1.0)
                for (i, is_const) in zip(range(len(u_parents)), self.is_constant)
                if i != index
            ]
            m2 = self._compute_message(m[2], mask, *alphas,
                                       ndim=0,
                                       plates_from=self.plates,
                                       plates_to=parent.plates)
            m3 = self._compute_message(m[3], mask,
                                       ndim=0,
                                       plates_from=self.plates,
                                       plates_to=parent.plates)

            msg = msg + [m2, m3]

        return msg


def Dot(*args, **kwargs):
    """
    Node for computing inner product of several Gaussian vectors.

    This is a simple wrapper of the much more general SumMultiply. For now, it
    is here for backward compatibility.
    """
    einsum = 'i' + ',i'*(len(args)-1)
    return SumMultiply(einsum, *args, **kwargs)



######################################################################
# Copyright (C) 2015 Jaakko Luttinen
#
# This file is licensed under the MIT License.
######################################################################


"""
Module for Bernoulli using the logistic function for Gaussian
"""

import numpy as np

from .node import ensureparents
from .expfamily import (ExponentialFamily,
                        useconstructor)
from .multinomial import (MultinomialMoments,
                          MultinomialDistribution,
                          Multinomial)
from .dirichlet import DirichletMoments
from .gaussian import GaussianMoments

from bayespy.utils import random
from bayespy.utils import misc


class CategoricalDistribution(MultinomialDistribution):
    """
    Class for the VMP formulas of categorical variables.
    """

    def __init__(self, categories):
        """
        Create VMP formula node for a categorical variable

        `categories` is the total number of categories.
        """
        if not isinstance(categories, int):
            raise ValueError("Number of categories must be integer")
        if categories < 0:
            raise ValueError("Number of categoriess must be non-negative")
        self.D = categories
        super().__init__(1)


    def compute_message_to_parent(self, parent, index, u, u_p):
        """
        Compute the message to a parent node.
        """
        return super().compute_message_to_parent(parent, index, u, u_p)


    def compute_phi_from_parents(self, u_p, mask=True):
        """
        Compute the natural parameter vector given parent moments.
        """
        return super().compute_phi_from_parents(u_p, mask=mask)


    def compute_moments_and_cgf(self, phi, mask=True):
        """
        Compute the moments and :math:`g(\phi)`.
        """
        return super().compute_moments_and_cgf(phi, mask=mask)


    def compute_cgf_from_parents(self, u_p):
        """
        Compute :math:`\mathrm{E}_{q(p)}[g(p)]`
        """
        return super().compute_cgf_from_parents(u_p)


    def compute_fixed_moments_and_f(self, x, mask=True):
        """
        Compute the moments and :math:`f(x)` for a fixed value.
        """

        # Check the validity of x
        x = np.asanyarray(x)
        if not misc.isinteger(x):
            raise ValueError("Values must be integers")
        if np.any(x < 0) or np.any(x >= self.D):
            raise ValueError("Invalid category index")

        # Form a binary matrix with only one non-zero (1) in the last axis
        u0 = np.zeros((np.size(x), self.D))
        u0[[np.arange(np.size(x)), np.ravel(x)]] = 1
        u0 = np.reshape(u0, np.shape(x) + (self.D,))
        u = [u0]

        # f(x) is zero
        f = 0

        return (u, f)


    def random(self, *phi, plates=None):
        """
        Draw a random sample from the distribution.
        """
        logp = phi[0]
        logp -= np.amax(logp, axis=-1, keepdims=True)
        p = np.exp(logp)
        return random.categorical(p, size=plates)


class Logistic(ExponentialFamily):
    r"""

    :cite:`Jaakkola:2000`

    The true probability density function:

    .. math::

       p(z=1|x) = g(x)
       \\
       p(z=0|x) = g(-x)

    which can be written as:

    .. math::

       p(z|x) = g(H_z)

    where :math:`H_z=(2z-1)x` and :math:`g(x)` is the logistic function:

    .. math::

       g(x) = \frac{1} {1 + e^{-x}}

    The log of the logistic function:

    .. math::

       \log g(x) = -\log(1 + e^{-x}) = \frac{x}{2} - \log(e^{x/2} + e^{-x/2})

    The latter term:

    .. math::

       f(x) = -\log(e^{x/2} + e^{-x/2})

    This is a convex function in the variable :math:`x^2`, thus it can be
    bounded globally with a first order Taylor expansion in the variable x^2:

    .. math::

       f(x) &\geq f(\xi) + \frac {\partial f(\xi)}{\partial(\xi^2)} (x^2 -
       \xi^2)
       \\
       &= -\frac{\xi}{2} + \log g(\xi) + \frac{1}{4\xi}\tanh(\frac{\xi}{2}) (x^2
       - \xi^2)

    Thus, the variational lower bound for the probability density function is:

    .. math::

       p(z|x) \geq g(xi) \exp( \frac{H_z-\xi}{2} - \lambda(\xi) (H_z^2 - \xi^2))

    and in log form:

    .. math::

       \log p(z|x) \geq \log g(xi) + ( \frac{H_z-\xi}{2} - \lambda(\xi) (H_z^2 -
       \xi^2) )

    where

    .. math::

       \lambda(\xi) = \frac {\tanh(\xi/2)} {4\xi}

    Now, this log lower bound is quadratic with respect to :math:`H_z`, thus it
    is quadratic with respect to :math:`x` and it is conjugate with the Gaussian
    distribution.  Re-organize terms:

    .. math::

       \log p(z|x) &\geq -\lambda(\xi)(2z-1)^2 x^2 + zx - \frac{1}{2}x -
       \frac{1}{2}\xi + \lambda(\xi) \xi^2 + \log g(\xi)
       \\
       &= -\lambda(\xi)(2z+1) x^2 + zx - \frac{1}{2}x -
       \frac{1}{2}\xi + \lambda(\xi) \xi^2 + \log g(\xi)
       \\
       &= z (-2\lambda(\xi) x^2 + x) - \lambda(\xi) x^2 - \frac{1}{2}x -
       \frac{1}{2}\xi + \lambda(\xi) \xi^2 + \log g(\xi)

    where we have used :math:`z^2=z`.

    See also
    --------
    Bernoulli, GaussianARD
    """


    _parent_moments = (
        GaussianMoments(()),
    )


    def __init__(self, x, **kwargs):
        """
        """
        super().__init__(x, **kwargs)


    @classmethod
    @ensureparents
    def _constructor(cls, x, **kwargs):
        """
        Constructs distribution and moments objects.
        """

        # Get the number of categories
        D = p.dims[0][0]

        parents = [p]
        moments = CategoricalMoments(D)
        distribution = CategoricalDistribution(D)

        return (parents,
                kwargs,
                ( (D,), ),
                cls._total_plates(kwargs.get('plates'),
                                  distribution.plates_from_parent(0, p.plates)),
                distribution,
                moments,
                cls._parent_moments)


    def __str__(self):
        """
        Print the distribution using standard parameterization.
        """
        raise NotImplementedError()



################################################################################
# Copyright (C) 2014 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
Module for the multinomial distribution node.
"""

import numpy as np
from scipy import special

from .expfamily import ExponentialFamily
from .expfamily import ExponentialFamilyDistribution
from .expfamily import useconstructor
from .dirichlet import Dirichlet, DirichletMoments
from .node import Moments, ensureparents

from bayespy.utils import random
from bayespy.utils import misc
from bayespy.utils import linalg


class MultinomialMoments(Moments):
    """
    Class for the moments of multinomial variables.
    """


    def __init__(self, categories):
        self.categories = categories
        self.dims = ( (categories,), )


    def compute_fixed_moments(self, x):
        """
        Compute the moments for a fixed value

        `x` must be a vector of counts.
        """

        # Check that counts are valid
        x = np.asanyarray(x)
        if not misc.isinteger(x):
            raise ValueError("Counts must be integer")
        if np.any(x < 0):
            raise ValueError("Counts must be non-negative")

        # Moments is just the counts vector
        u0 = x.copy()
        return [u0]


    @classmethod
    def from_values(cls, x):
        D = np.shape(x)[-1]
        return cls( (D,) )


class MultinomialDistribution(ExponentialFamilyDistribution):
    """
    Class for the VMP formulas of multinomial variables.
    """


    def __init__(self, trials):
        """
        Create VMP formula node for a multinomial variable

        `trials` is the total number of trials.
        """
        trials = np.asanyarray(trials)
        if not misc.isinteger(trials):
            raise ValueError("Number of trials must be integer")
        if np.any(trials < 0):
            raise ValueError("Number of trials must be non-negative")
        self.N = trials
        super().__init__()


    def compute_message_to_parent(self, parent, index, u, u_p):
        """
        Compute the message to a parent node.
        """
        if index == 0:
            return [ u[0].copy() ]
        else:
            raise ValueError("Index out of bounds")


    def compute_phi_from_parents(self, u_p, mask=True):
        """
        Compute the natural parameter vector given parent moments.
        """
        logp = u_p[0]
        return [logp]


    def compute_moments_and_cgf(self, phi, mask=True):
        r"""
        Compute the moments and :math:`g(\phi)`.

        .. math::

           \overline{\mathbf{u}}
           = \mathrm{E}[x]
           = N \cdot \begin{bmatrix}
             \frac{e^{\phi_1}}{\sum_i e^{\phi_i}}
             & \cdots &
             \frac{e^{\phi_D}}{\sum_i e^{\phi_i}} \end{bmatrix}
        """
        # Compute the normalized probabilities in a numerically stable way
        (p, logsum_p) = misc.normalized_exp(phi[0])
        N = np.expand_dims(self.N, -1)
        u0 = N * p
        u = [u0]
        g = -np.squeeze(N * logsum_p, axis=-1)
        return (u, g)


    def compute_cgf_from_parents(self, u_p):
        r"""
        Compute :math:`\mathrm{E}_{q(p)}[g(p)]`
        """
        return 0


    def compute_fixed_moments_and_f(self, x, mask=True):
        r"""
        Compute the moments and :math:`f(x)` for a fixed value.
        """

        # Check that counts are valid
        x = np.asanyarray(x)
        if not misc.isinteger(x):
            raise ValueError("Counts must be integers")
        if np.any(x < 0):
            raise ValueError("Counts must be non-negative")
        if np.any(np.sum(x, axis=-1) != self.N):
            raise ValueError("Counts must sum to the number of trials")

        # Moments is just the counts vector
        u0 = x.copy()
        u = [u0]

        f = special.gammaln(self.N+1) - np.sum(special.gammaln(x+1), axis=-1)

        return (u, f)


    def random(self, *phi, plates=None):
        r"""
        Draw a random sample from the distribution.
        """
        (p, _) = misc.normalized_exp(phi[0])
        return random.multinomial(self.N, p, size=plates)


    def compute_gradient(self, g, u, phi):
        r"""
        Compute the Euclidean gradient.

        In order to compute the Euclidean gradient, we first need to derive the
        gradient of the moments with respect to the variational parameters:

        .. math::

           \mathrm{d}\overline{u}_i
           = N \cdot \frac {e^{\phi_i} \mathrm{d}\phi_i \sum_j e^{\phi_j}}
                           {(\sum_k e^{\phi_k})^2}
             - N \cdot \frac {e^{\phi_i} \sum_j e^\phi_j \mathrm{d}\phi_j}
                             {(\sum_k e^{\phi_k})^2}
           = \overline{u}_i \mathrm{d}\phi_i
             - \overline{u}_i \sum_j \frac{\overline{u}_j}{N} \mathrm{d}\phi_j


        Now we can make use of the chain rule. Given the Riemannian gradient
        :math:`\tilde{\nabla}` of the variational lower bound
        :math:`\mathcal{L}` with respect to the variational parameters
        :math:`\phi`, put the above result to the derivative term and
        re-organize the terms to get the Euclidean gradient :math:`\nabla`:

        .. math::

           \mathrm{d}\mathcal{L}
           = \tilde{\nabla}^T \mathrm{d}\overline{\mathbf{u}}
           = \sum_i \tilde{\nabla}_i \mathrm{d}\overline{u}_i
           = \sum_i \tilde{\nabla}_i (
                 \overline{u}_i \mathrm{d}\phi_i
                 - \overline{u}_i \sum_j \frac {\overline{u}_j} {N} \mathrm{d}\phi_j
             )
           = \sum_i \left(\tilde{\nabla}_i \overline{u}_i \mathrm{d}\phi_i
             - \frac{\overline{u}_i}{N} \mathrm{d}\phi_i \sum_j \tilde{\nabla}_j \overline{u}_j \right)
           \equiv \nabla^T \mathrm{d}\phi

        Thus, the Euclidean gradient is:

        .. math::

           \nabla_i = \tilde{\nabla}_i \overline{u}_i - \frac{\overline{u}_i}{N}
                      \sum_j \tilde{\nabla}_j \overline{u}_j

        See also
        --------

        compute_moments_and_cgf : Computes the moments
            :math:`\overline{\mathbf{u}}` given the variational parameters
            :math:`\phi`.

        """
        return u[0] * (g - linalg.inner(g, u[0])[...,None] / self.N)

    def squeeze(self, axis):
        try:
            N_squeezed = np.squeeze(self.N, axis)
        except ValueError as err:
            raise ValueError(
                "The number of trials must be constant over a squeezed axis, "
                "so the corresponding array axis must be singleton. "
                "Cannot squeeze axis {0} from a multinomial distribution "
                "because the number of trials arrays has shape {2}, so "
                "the given axis has length {1} != 1. ".format(
                    axis,
                    np.shape(self.N)[axis],
                    np.shape(self.N),
                )
            ) from err
        else:
            return MultinomialDistribution(N_squeezed)


class Multinomial(ExponentialFamily):
    r"""
    Node for multinomial random variables.

    Assume there are :math:`K` categories and :math:`N` trials each of which
    leads a success for exactly one of the categories.  Given the probabilities
    :math:`p_0,\ldots,p_{K-1}` for the categories, multinomial distribution is
    gives the probability of any combination of numbers of successes for the
    categories.

    The node models the number of successes :math:`x_k \in \{0, \ldots, n\}` in
    :math:`n` trials with probability :math:`p_k` for success in :math:`K`
    categories.

    .. math::

        \mathrm{Multinomial}(\mathbf{x}| N, \mathbf{p}) = \frac{N!}{x_0!\cdots
        x_{K-1}!} p_0^{x_0} \cdots p_{K-1}^{x_{K-1}}

    Parameters
    ----------

    n : scalar or array
        :math:`N`, number of trials
    p : Dirichlet-like node or (...,K)-array
        :math:`\mathbf{p}`, probabilities of successes for the categories

    See also
    --------

    Dirichlet, Binomial, Categorical
    """


    def __init__(self, n, p, **kwargs):
        """
        Create Multinomial node.
        """
        super().__init__(n, p, **kwargs)


    @classmethod
    def _constructor(cls, n, p, **kwargs):
        """
        Constructs distribution and moments objects.

        This method is called if useconstructor decorator is used for __init__.

        Becase the distribution and moments object depend on the number of
        categories, that is, they depend on the parent node, this method can be
        used to construct those objects.
        """

        # Get the number of categories
        p = cls._ensure_moments(p, DirichletMoments)
        D = p.dims[0][0]

        moments = MultinomialMoments(D)
        parent_moments = (p._moments,)

        parents = [p]

        distribution = MultinomialDistribution(n)

        return (parents,
                kwargs,
                moments.dims,
                cls._total_plates(kwargs.get('plates'),
                                  distribution.plates_from_parent(0, p.plates),
                                  np.shape(n)),
                distribution,
                moments,
                parent_moments)


    def __str__(self):
        """
        Print the distribution using standard parameterization.
        """
        logsum_p = misc.logsumexp(self.phi[0], axis=-1, keepdims=True)
        p = np.exp(self.phi[0] - logsum_p)
        p /= np.sum(p, axis=-1, keepdims=True)
        return ("%s ~ Multinomial(p)\n"
                "  p = \n"
                "%s\n"
                % (self.name, p))



################################################################################
# Copyright (C) 2014-2015 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
Demonstrate Riemannian conjugate gradient
"""

import numpy as np

from bayespy.nodes import (GaussianARD,
                           Gamma,
                           SumMultiply)

from bayespy.utils import random

from bayespy.inference.vmp.vmp import VB

import bayespy.plot as bpplt

from bayespy.demos import mog


def pca():

    np.random.seed(41)

    M = 10
    N = 3000
    D = 5

    # Construct the PCA model
    alpha = Gamma(1e-3, 1e-3, plates=(D,), name='alpha')
    W = GaussianARD(0, alpha, plates=(M,1), shape=(D,), name='W')
    X = GaussianARD(0, 1, plates=(1,N), shape=(D,), name='X')
    tau = Gamma(1e-3, 1e-3, name='tau')
    W.initialize_from_random()
    F = SumMultiply('d,d->', W, X)
    Y = GaussianARD(F, tau, name='Y')

    # Observe data
    data = np.sum(np.random.randn(M,1,D-1) * np.random.randn(1,N,D-1), axis=-1) + 1e-1 * np.random.randn(M,N)
    Y.observe(data)

    # Initialize VB engine
    Q = VB(Y, X, W, alpha, tau)

    # Take one update step (so phi is ok)
    Q.update(repeat=1)
    Q.save()

    # Run VB-EM
    Q.update(repeat=200)
    bpplt.pyplot.plot(np.cumsum(Q.cputime), Q.L, 'k-')

    # Restore the state
    Q.load()

    # Run Riemannian conjugate gradient
    #Q.optimize(X, alpha, maxiter=100, collapsed=[W, tau])
    Q.optimize(W, tau, maxiter=100, collapsed=[X, alpha])
    bpplt.pyplot.plot(np.cumsum(Q.cputime), Q.L, 'r:')

    bpplt.pyplot.show()


def mixture_of_gaussians():
    """Collapsed Riemannian conjugate gradient demo

    This is similar although not exactly identical to an experiment in
    (Hensman et al 2012).
    """

    np.random.seed(41)

    # Number of samples
    N = 1000
    # Number of clusters in the model (five in the data)
    K = 10

    # Overlap parameter of clusters
    R = 2

    # Construct the model
    Q = mog.gaussianmix_model(N, K, 2, covariance='diagonal')

    # Generate data from five Gaussian clusters
    mu = np.array([[0, 0],
                   [R, R],
                   [-R, R],
                   [R, -R],
                   [-R, -R]])
    Z = random.categorical(np.ones(5), size=N)
    data = np.empty((N, 2))
    for n in range(N):
        data[n,:] = mu[Z[n]] + np.random.randn(2)
    Q['Y'].observe(data)

    # Take one update step (so phi is ok)
    Q.update(repeat=1)
    Q.save()

    # Run standard VB-EM
    Q.update(repeat=1000, tol=0)
    bpplt.pyplot.plot(np.cumsum(Q.cputime), Q.L, 'k-')

    # Restore the initial state
    Q.load()

    # Run Riemannian conjugate gradient
    Q.optimize('alpha', 'X', 'Lambda', collapsed=['z'], maxiter=300, tol=0)
    bpplt.pyplot.plot(np.cumsum(Q.cputime), Q.L, 'r:')

    bpplt.pyplot.xlabel('CPU time (in seconds)')
    bpplt.pyplot.ylabel('VB lower bound')
    bpplt.pyplot.legend(['VB-EM', 'Collapsed Riemannian CG'], loc='lower right')

    ## bpplt.pyplot.figure()
    ## bpplt.pyplot.plot(data[:,0], data[:,1], 'rx')
    ## bpplt.pyplot.title('Data')

    bpplt.pyplot.show()
    

if __name__ == "__main__":
    #pca()
    mixture_of_gaussians()
